#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Plotting vector data.

This module can be used in two different ways:

1. As a library. Just import the module and call the functions.
   This is the way, how this module is used in openpivgui, for
   example.

2. As a terminal-application. Execute
   python3 -m openpivgui.vec_plot --help
   for more information.
   This is the way, how this module ist used in JPIV, for example.
   For now, not all functions are callable in this way.
"""

from matplotlib import pyplot as plt
import matplotlib
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.figure import Figure
from copy import copy
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import argparse
__licence__ = '''
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

__email__ = 'vennemann@fh-muenster.de'


# creating a custom rainbow colormap

# creating a custom rainbow colormap
short_rainbow = {'red': ((0.0, 0.0, 0.0),
                         (0.2, 0.2, 0.2),
                         (0.5, 0.0, 0.0),
                         (0.8, 1.0, 1.0),
                         (1.0, 1.0, 1.0)),
                 'green': ((0.0, 0.0, 0.0),
                           (0.2, 1.0, 1.0),
                           (0.5, 1.0, 1.0),
                           (0.8, 1.0, 1.0),
                           (1.0, 0.0, 0.0)),
                 'blue': ((0.0, 1.0, 1.0),
                          (0.2, 1.0, 1.0),
                          (0.5, 0.0, 0.0),
                          (0.8, 0.0, 0.0),
                          (1.0, 0.0, 0.0))}

long_rainbow = {'red': ((0.0, 0.0, 0.0),
                        (0.1, 0.5, 0.5),
                        (0.2, 0.0, 0.0),
                        (0.3, 0.2, 0.2),
                        (0.5, 0.0, 0.0),
                        (0.7, 1.0, 1.0),
                        (0.8, 1.0, 1.0),
                        (1.0, 1.0, 1.0)),
                'green': ((0.0, 0.0, 0.0),
                          (0.1, 0.0, 0.0),
                          (0.2, 0.0, 0.0),
                          (0.3, 1.0, 1.0),
                          (0.5, 1.0, 1.0),
                          (0.7, 1.0, 1.0),
                          (0.8, 0.0, 0.0),
                          (1.0, 0.3, 0.3)),
                'blue': ((0.0, 0.0, 0.0),
                         (0.1, 0.5, 0.5),
                         (0.2, 1.0, 1.0),
                         (0.3, 1.0, 1.0),
                         (0.5, 0.0, 0.0),
                         (0.7, 0.0, 0.0),
                         (0.8, 0.0, 0.0),
                         (1.0, 1.0, 1.0))}

short_rainbow = LinearSegmentedColormap('my_colormap', short_rainbow, 256)
long_rainbow = LinearSegmentedColormap('my_colormap', long_rainbow, 256)


def histogram(data, figure, quantity, bins, log_y):
    """
        Plot an histogram.

        Plots an histogram of the specified quantity.

        Parameters
        ----------
        data : pandas.DataFrame
            Data to plot.
        figure : matplotlib.figure.Figure
            An (empty) Figure object.
        quantity : str
            Either v (abs v), v_x (x-component) or v_y (y-component).
        bins : int
             Number of bins (bars) in the histogram.
        log_scale : boolean
            Use logaritmic vertical axis.
    """

    if quantity == 'v':
        xlabel = 'absolute displacement'
        h_data = np.array([(l[2]**2 + l[3]**2)**0.5 for l in data])
    elif quantity == 'v_x':
        xlabel = 'x displacement'
        h_data = np.array([l[2] for l in data])
    elif quantity == 'v_y':
        xlabel = 'y displacement'
        h_data = np.array([l[3] for l in data])
    ax = figure.add_subplot(111)
    if log_y:
        ax.set_yscale("log")
    ax.hist(h_data, bins, label=quantity)
    ax.set_xlabel(xlabel)
    ax.set_ylabel('number of vectors')
    ax.set_title(parameter['plot_title'])


def profiles(data, parameter, fname, figure, orientation):
    """
        Plot velocity profiles.

        Line plots of the velocity component specified.

        Parameters
        ----------
        data : pandas.DataFrame
            Data to plot.
        fname : str
            A filename containing vector data.
            (will be deprecated in later updates)
        figure : matplotlib.figure.Figure
            An (empty) Figure object.
        orientation : str
            horizontal: Plot v_y over x.
            vertical: Plot v_x over y.
    """
    #data = data.to_numpy().astype(np.float)
    data = np.loadtxt(fname)

    dim_x, dim_y = get_dim(data)

    p_data = []

    ax = figure.add_subplot(111)

    if orientation == 'horizontal':
        xlabel = 'x position'
        ylabel = 'y displacement'

        for i in range(0, dim_y, parameter['profiles_jump']):
            p_data.append(data[dim_x * i:dim_x * (i + 1), 3])
        # print(p_data[-1])
        for p in p_data:
            # print(len(p))
            ax.plot(range(dim_x), p, '.-')

    elif orientation == 'vertical':
        xlabel = 'y position'
        ylabel = 'x displacement'

        for i in range(0, dim_x, parameter['profiles_jump']):
            p_data.append(data[i::dim_x, 2])

        for p in p_data:
            ax.plot(range(dim_y), p, '.-')

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(parameter['plot_title'])


def scatter(data, figure):
    """
        Scatter plot.

        Plots v_y over v_x.

        Parameters
        ----------
        data : pandas.DataFrame
            Data to plot.
        figure : matplotlib.figure.Figure
            An (empty) Figure object.
    """
    data = data.to_numpy()

    v_x = data[:, 2]
    v_y = data[:, 3]

    ax = figure.add_subplot(111)

    ax.scatter(v_x, v_y, label='scatter')

    ax.set_xlabel('x displacement')
    ax.set_ylabel('y displacement')


def vector(data, parameter, figure, invert_yaxis=True, valid_color='blue',
           invalid_color='red', **kw):
    """
        Display a vector plot.

        Parameters
        ----------
        data : pandas.DataFrame
            Data to plot.
        figure : matplotlib.figure.Figure
            An (empty) Figure object.
    """
    data = data.to_numpy().astype(np.float)

    try:
        invalid = data[:, 4].astype('bool')
    except BaseException:
        invalid = np.asarray([True for i in range(len(data))])

    # tilde means invert:
    valid = ~invalid

    ax = figure.add_subplot(111)

    ax.quiver(data[invalid, 0],
              data[invalid, 1],
              data[invalid, 2],
              data[invalid, 3],
              color=invalid_color,
              label='invalid', **kw)

    ax.quiver(data[valid, 0],
              data[valid, 1],
              data[valid, 2],
              data[valid, 3],
              color=valid_color,
              label='valid', **kw)

    if invert_yaxis:
        for ax in figure.get_axes():
            ax.invert_yaxis()

    ax.set_xlabel('x position')
    ax.set_ylabel('y position')
    ax.set_title(parameter['plot_title'])


def contour(data, parameter, figure):
    '''Display a contour plot

    Parameters
    ----------
    data : pandas.DataFrame
        Data to plot.
    parameter : openpivgui.OpenPivParams.py
        Parameter-object.
    figure : matplotlib.figure.Figure
       An (empty) Figure object.
    '''
    # figure for subplot
    ax = figure.add_subplot(111)
    # iteration to set value types to float
    for i in list(data.columns.values):
        data[i] = data[i].astype(float)
    # choosing velocity for the colormap and add it to an new colummn in data
    if parameter['velocity_color'] == 'vx':
        data['abs'] = data.vx
    elif parameter['velocity_color'] == 'vy':
        data['abs'] = data.vy
    else:
        data['abs'] = (data.vx**2 + data.vy**2)**0.5
    # pivot table for contour function
    data_pivot = data.pivot(index='y',
                            columns='x',
                            values='abs')
    # try to get limits, if not possible set to None
    try:
        vmin = float(parameter['vmin'])
    except BaseException:
        vmin = None
    try:
        vmax = float(parameter['vmax'])
    except BaseException:
        vmax = None
    # settings for color scheme of the contour plot
    if vmax is not None and vmin is not None:
        levels = np.linspace(vmin, vmax, int(parameter['color_levels']))
    elif vmax is not None:
        levels = np.linspace(0, vmax, int(parameter['color_levels']))
    elif vmin is not None:
        vmax = data_pivot.max().max()
        levels = np.linspace(vmin, vmax, int(parameter['color_levels']))
    else:
        levels = int(parameter['color_levels'])
    # Choosing the correct colormap
    if parameter['color_map'] == 'short rainbow':
        colormap = short_rainbow
    elif parameter['color_map'] == 'long rainbow':
        colormap = long_rainbow
    else:
        colormap = parameter['color_map']
    # set contour plot to the variable fig to add a colorbar
    if parameter['extend_cbar']:
        extend = 'both'
    else:
        extend = None
    fig = ax.contourf(data_pivot.columns,
                      data_pivot.index,
                      data_pivot.values,
                      levels=levels,
                      cmap=colormap,
                      vmin=vmin,
                      vmax=vmax,
                      extend=extend)

    # set the colorbar to the variable cb to add a description
    cb = plt.colorbar(fig, ax=ax)

    # set origin to top left or bottom left
    if parameter['invert_yaxis']:
        ax.set_ylim(ax.get_ylim()[::-1])

    # description to the contour lines
    cb.ax.set_ylabel('Velocity [m/s]')

    # labels for the axes
    ax.set_xlabel('x-position')
    ax.set_ylabel('y-position')

    # plot title from the GUI
    ax.set_title(parameter['plot_title'])


def contour_and_vector(data, parameter, figure, **kw):
    """
        Display a contour plot

        Parameters
        ----------
        data : pandas.DataFrame
            Data to plot.
        parameter : openpivgui.OpenPivParams.py
            Parameter-object.
        figure : matplotlib.figure.Figure
           An (empty) Figure object.
    """
    # figure for subplot
    ax = figure.add_subplot(111)
    # iteration to set value types to float
    for i in list(data.columns.values):
        data[i] = data[i].astype(float)
    # choosing velocity for the colormap and add it to an new colummn in data
    if parameter['velocity_color'] == 'vx':
        data['abs'] = data.vx
    elif parameter['velocity_color'] == 'vy':
        data['abs'] = data.vy
    else:
        data['abs'] = (data.vx**2 + data.vy**2)**0.5
    # pivot table for contour function
    data_pivot = data.pivot(index='y',
                            columns='x',
                            values='abs')
    # try to get limits, if not possible set to None
    try:
        vmin = float(parameter['vmin'])
    except BaseException:
        vmin = None
    try:
        vmax = float(parameter['vmax'])
    except BaseException:
        vmax = None
    # settings for color scheme of the contour plot
    if vmax is not None and vmin is not None:
        levels = np.linspace(vmin, vmax, int(parameter['color_levels']))
    elif vmax is not None:
        levels = np.linspace(0, vmax, int(parameter['color_levels']))
    elif vmin is not None:
        vmax = data_pivot.max().max()
        levels = np.linspace(vmin, vmax, int(parameter['color_levels']))
    else:
        levels = int(parameter['color_levels'])
    # Choosing the correct colormap
    if parameter['color_map'] == 'short rainbow':
        colormap = short_rainbow
    elif parameter['color_map'] == 'long rainbow':
        colormap = long_rainbow
    else:
        colormap = parameter['color_map']

    # set contour plot to the variable fig to add a colorbar
    if parameter['extend_cbar']:
        extend = 'both'
    else:
        extend = None
    fig = ax.contourf(data_pivot.columns,
                      data_pivot.index,
                      data_pivot.values,
                      levels=levels,
                      cmap=colormap,
                      vmin=vmin,
                      vmax=vmax,
                      extend=extend)

    # quiver plot
    data = data.to_numpy().astype(np.float)

    try:
        invalid = data[:, 4].astype('bool')
    except BaseException:
        invalid = np.asarray([True for i in range(len(data))])

    # tilde means invert:
    valid = ~invalid

    ax.quiver(data[invalid, 0],
              data[invalid, 1],
              data[invalid, 2],
              data[invalid, 3],
              color=parameter['invalid_color'],
              label='invalid', **kw)

    ax.quiver(data[valid, 0],
              data[valid, 1],
              data[valid, 2],
              data[valid, 3],
              color=parameter['valid_color'],
              label='valid', **kw)

    # set the colorbar to the variable cb to add a description
    cb = plt.colorbar(fig, ax=ax)

    # set origin to top left or bottom left
    if parameter['invert_yaxis']:
        ax.set_ylim(ax.get_ylim()[::-1])

    # description to the contour lines
    cb.ax.set_ylabel('Velocity [m/s]')

    # labels for the axes
    ax.set_xlabel('x-position')
    ax.set_ylabel('y-position')

    # plot title from the GUI
    ax.set_title(parameter['plot_title'])


def streamlines(data, parameter, figure):
    """
        Display a streamline plot.

        Parameters
        ----------
        data : pandas.DataFrame
            Data to plot.
        parameter : openpivgui.OpenPivParams.py
            Parameter object.
        figure : matplotlib.figure.Figure
            An (empty) Figure object.
    """
    ax = figure.add_subplot(111)

    # make sure all values are from type float
    for i in list(data.columns.values):
        data[i] = data[i].astype(float)

    # get density for streamline plot.
    try:
        density = (float(list(parameter['streamline_density'].split(','))[0]),
                   float(list(parameter['streamline_density'].split(','))[1]))
    except BaseException:
        density = float(parameter['streamline_density'])

    # Choosing the correct colormap
    if parameter['color_map'] == 'short rainbow':
        colormap = short_rainbow
    elif parameter['color_map'] == 'long rainbow':
        colormap = long_rainbow
    else:
        colormap = parameter['color_map']

    # pivot table for streamline plot
    data_vx = data.pivot(index='y',
                         columns='x',
                         values='vx')
    data_vy = data.pivot(index='y',
                         columns='x',
                         values='vy')

    # choosing data for the colormap
    if parameter['velocity_color'] == 'vx':
        color_values = data_vx.values
    elif parameter['velocity_color'] == 'vy':
        color_values = data_vy.values
    else:
        color_values = (data_vx.values**2 + data_vy.values**2)**0.5

    # try to create streamline plot. If values are not equally spaced the
    # exception will space the values equally (mean difference is
    # calculated.)

    try:
        fig = ax.streamplot(data_vx.columns,
                            data_vx.index,
                            data_vx.values,
                            data_vy.values,
                            density=density,
                            color=color_values,
                            cmap=colormap,
                            integration_direction=parameter['integrate_dir'],
                            linewidth=parameter['vec_width'])
    except BaseException:
        # get dimension of the DataFrame
        dim = [len(set(data.x)), len(set(data.y))]

        # calculate mean difference for x and y values
        diff = [round(np.mean(
            [data.x[i + 1] - data.x[i] for i in range(dim[0] - 1)]), 6),
            round(np.mean([data.y[dim[0] * (i + 1)] - data.y[dim[0] * i]
                           for i in range(dim[1] - 1)]), 6)]

        # this list is initialized with starting values and will be added by
        # equally spaced values.
        cache = [round(copy(data.x[0]), 6), round(copy(data.y[0]), 6)]

        # nested lists with equally spaced coordinates
        coordinates = [[], []]

        # loop for calculating the new x data
        j = 1
        for i in range(1, len(data)):
            if i == dim[0] * j:
                coordinates[0].append(round(cache[0], 6))
                cache[0] = coordinates[0][0]
                j += 1
            else:
                coordinates[0].append(round(cache[0], 6))
                cache[0] += diff[0]
        coordinates[0].append(round(cache[0], 6))

        # loop for calculating the new y data
        j = 1
        for i in range(len(data)):
            if i == dim[0] * j:
                cache[1] += diff[1]
                coordinates[1].append(round(cache[1], 6))
                j += 1
            else:
                coordinates[1].append(round(cache[1], 6))

        # overwrite the old x and y values with the new ones
        data.x = coordinates[0]
        data.y = coordinates[1]

        # create new pivot tables for streamline plot
        data_vx = data.pivot(index='y', columns='x', values='vx')
        data_vy = data.pivot(index='y', columns='x', values='vy')

        # choosing data for the colormap
        if parameter['velocity_color'] == 'vx':
            color_values = data_vx.values
        elif parameter['velocity_color'] == 'vy':
            color_values = data_vy.values
        else:
            color_values = (data_vx.values**2 + data_vy.values**2)**0.5

        # new streamline plot with equally spaced coordinates
        fig = ax.streamplot(data_vx.columns,
                            data_vx.index,
                            data_vx.values,
                            data_vy.values,
                            density=density,
                            color=color_values,
                            cmap=colormap,
                            integration_direction=parameter['integrate_dir'],
                            linewidth=parameter['vec_width'])
    # add colorbar
    cb = plt.colorbar(fig.lines, ax=ax)
    cb.ax.set_ylabel('Velocity [m/s]')

    # set origin to top left or bottom left
    if parameter['invert_yaxis']:
        ax.set_ylim(ax.get_ylim()[::-1])

    # add diagram options
    ax.set_xlabel('x-position')
    ax.set_ylabel('y-position')
    ax.set_title(parameter['plot_title'])


def pandas_plot(data, parameter, figure):
    """
        Display a plot with the pandas plot utility.

        Parameters
        ----------
        data : pandas.DataFrame
            Data to plot.
        parameter : openpivgui.OpenPivParams.py
            Parameter-object.
        figure : matplotlib.figure.Figure
            An (empty) figure.

        Returns
        -------
        None.

    """
    # set boolean for chosen axis scaling
    if parameter['plot_scaling'] == 'None':
        logx, logy, loglog = False, False, False
    elif parameter['plot_scaling'] == 'logx':
        logx, logy, loglog = True, False, False
    elif parameter['plot_scaling'] == 'logy':
        logx, logy, loglog = False, True, False
    elif parameter['plot_scaling'] == 'loglog':
        logx, logy, loglog = False, False, True
    # add subplot
    ax = figure.add_subplot(111)
    # set limits initially to None
    xlim = None
    ylim = None
    # try to set limits, if not possible (no entry) --> None
    try:
        xlim = (float(list(parameter['plot_xlim'].split(','))[0]),
                float(list(parameter['plot_xlim'].split(','))[1]))
    except BaseException:
        pass
        #print('No Values or wrong syntax for x-axis limitation.')
    try:
        ylim = (float(list(parameter['plot_ylim'].split(','))[0]),
                float(list(parameter['plot_ylim'].split(','))[1]))
    except BaseException:
        pass
        #print('No Values or wrong syntax for y-axis limitation.')
    # iteration to set value types to float
    for i in list(data.columns.values):
        data[i] = data[i].astype(float)

    if parameter['plot_type'] == 'histogram':
        # get column names as a list for comparing with chosen histogram
        # quantity
        col_names = list(data.columns.values)
        # if loop for histogram quantity
        if parameter['histogram_quantity'] == 'v_x':
            data_hist = data[col_names[2]]
        elif parameter['histogram_quantity'] == 'v_y':
            data_hist = data[col_names[3]]
        elif parameter['histogram_quantity'] == 'v':
            data_hist = (data[col_names[2]]**2 + data[col_names[3]]**2)**0.5
        # histogram plot
        ax.hist(data_hist,
                bins=int(parameter['histogram_bins']),
                label=parameter['histogram_quantity'],
                log=logy,
                range=xlim,
                density=parameter['histogram_normalize'],
                histtype=parameter['histogram_type'],
                )
        ax.grid(parameter['plot_grid'])
        ax.legend()
        ax.set_xlabel('velocity [m/s]')
        ax.set_ylabel('number of vectors')
        ax.set_title(parameter['plot_title'])
    else:
        data.plot(x=parameter['u_data'],
                  y=parameter['v_data'],
                  kind=parameter['plot_type'],
                  title=parameter['plot_title'],
                  grid=parameter['plot_grid'],
                  legend=parameter['plot_legend'],
                  logx=logx,
                  logy=logy,
                  loglog=loglog,
                  xlim=xlim,
                  ylim=ylim,
                  ax=ax)


def get_dim(array):
    '''Computes dimension of vector data.

    Assumes data to be organised as follows (example):
    x  y  v_x v_y ..
    16 16 4.5 3.2 ..
    32 16 4.3 3.1 ..
    16 32 4.2 3.5 ..
    32 32 4.5 3.2 ..
    .. .. ..  ..

    Parameters
    ----------
    array : np.array
        Flat numpy array.

    Returns
    -------
    tuple
        Dimension of the vector field (x, y).
    '''
    return(len(set(array[:, 0])),
           len(set(array[:, 1])))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Plot vector data.')
    parser.add_argument('--plot_type',
                        type=str,
                        required=False,
                        choices=['histogram',
                                 'profiles',
                                 'vector',
                                 'scatter',
                                 'contour'
                                 'contour_and_vector',
                                 'streamlines'],
                        default='vector',
                        help='type of plot')
    parser.add_argument('--fname',
                        required=True,
                        type=str,
                        help='name of vector data file')
    parser.add_argument('--quantity',
                        type=str,
                        required=False,
                        choices=['v', 'v_x', 'v_y'],
                        default='v',
                        help='quantity to plot')
    parser.add_argument('--bins',
                        type=int,
                        required=False,
                        default=20,
                        help='number of histogram bins')
    parser.add_argument('--log_y',
                        type=bool,
                        required=False,
                        default=False,
                        help='logarithmic y-axis')
    parser.add_argument('--orientation',
                        type=str,
                        required=False,
                        choices=['horizontal', 'vertical'],
                        default='vertical',
                        help='plot profiles, either horizontal ' +
                             '(v_y over x) or vertical (v_x over y)')
    parser.add_argument('--invert_yaxis',
                        type=str,
                        required=False,
                        default=True,
                        help='Invert y-axis of vector plot')
    args = parser.parse_args()
    data = np.loadtxt(args.fname)
    fig = Figure()
    if args.plot_type == 'histogram':
        histogram(data,
                  fig,
                  quantity=args.quantity,
                  bins=args.bins,
                  log_y=args.log_y)
    elif args.plot_type == 'profiles':
        profiles(data,
                 fig,
                 orientation=args.orientation)
    elif args.plot_type == 'vector':
        vector(data,
               fig,
               invert_yaxis=args.invert_yaxis)
    elif args.plot_type == 'scatter':
        scatter(data,
                fig)
    elif args.plot_type == 'contour':
        print('Not yet implemented')
    elif args.plot_type == 'contour_and_vector':
        print('Not yet implemented')
    elif args.plot_type == 'streamlines':
        print('Not yet implemented')
    plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''A class for simple parameter handling.

This class is also used as a basis for automated widget creation
by OpenPivGui.
'''

import os
import json

__licence__ = '''
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

__email__ = 'vennemann@fh-muenster.de'


class OpenPivParams:
    """
        A class for convenient parameter handling.

        Widgets are automatically created based on the content of the
        variables in the dictionary OpenPivParams.default.

        The entries in OpenPivParams.default are assumed to follow this
        pattern:

        (str) key:
            [(int) index,
             (str) type,
             value,
             (tuple) hints,
             (str) label,
             (str) help]

        The index is used for sorting and grouping, because Python
        dictionaries below version 3.7 do not preserve their order. A
        corresponding input widget is chosen based on the type string:

            None:                    no widget, no variable, but a rider
            boolean:                 checkbox
            str[]:                   listbox
            text:                    text area
            other (float, int, str): entry (if hints not None: option menu)

        A label is placed next to each input widget. The help string is
        displayed as a tooltip.

        The parameter value is directly accessible via indexing the base
        variable name. For example, if your OpenPivParams object variable
        name is »my_settings«, you can access a value by typing:

        my_settings[key]

        This is a shortcut for my_settings.param[key]. To access other
        fields, use my_settings.label[key], my_settings.help[key] and so on.
    """

    def __init__(self):
        # hard coded location of the parameter file in the home dir:
        self.params_fname = os.path.expanduser('~' + os.sep +
                                               'open_piv_gui.json')
        # grouping and sorting based on an index:
        self.GENERAL = 1000
        self.PREPROC = 2000
        self.PIVPROC = 3000
        self.VALIDATION = 6000
        self.POSTPROC = 7000
        self.PLOTTING = 8000
        self.LOGGING = 9000
        self.USER = 10000
        # remember the current file filter
        # (one of the comma separated values in ['navi_pattern']):
        self.navi_position = 0
        # these are the default parameters, basis for widget creation:
        self.default = {
            #########################################################
            # Place additional variables in the following sections. #
            # Widgets are created automatically. Don't care about   #
            # saving and restoring - new variables are included     #
            # automatically.                                        #
            #########################################################
            # general
            'general':
                [1000,
                 None,  # type None: This will create a rider.
                 None,
                 None,
                 'General',
                 None],

            'fnames':
                [1010,  # index, here: group GENERAL
                 'str[]',  # type
                 [],  # value
                 None,  # hint (used for option menu, if not None)
                 'filenames',  # label
                 None],  # help (tooltip)

            'general_frame':
                [1015, 'labelframe', None,
                 None,
                 'General settings',
                 None],

            'navi_pattern':
                [1110, 'sub',
                 'png$, tif$, bmp$, pgm$, vec$, ' +
                 r'DCC_[0-9]+\.vec$, ' +
                 r'FFT_[0-9]+\.vec$, ' +
                 r'sig2noise\.vec$, ' +
                 r'std_thrhld\.vec, ' +
                 r'med_thrhld\.vec, ' +
                 r'glob_thrhld\.vec, ' +
                 r'repl\.vec$, ' +
                 r'smthn\.vec$ ',
                 None,
                 'navigation pattern',
                 'Regular expression patterns for filtering the files ' +
                 'in the current directory. Use the back and forward ' +
                 'buttons to apply a different filter.'],

            'pandas_sub_frame':
                [1200, 'sub_labelframe', None,
                 None,
                 'Pandas',
                 None],

            'warnings': [1020, 'bool', 'True', None, 'Enable popup warnings',
                         'Enable popup warning messages (recommended).'],

            'pop_up_info': [1025, 'bool', 'True', None, 'Enable popup info',
                            'Enable popup information messages (recommended).'],

            'multicore_frame': [1030, 'sub_labelframe', None, None,
                                'multicore settings', None],

            'manual_select_cores': [1035, 'sub_bool', 'True', None,
                                    'manually select cores',
                                    'Mannualy select cores. If not seected, all ' +
                                    'available cores will be used.'],

            'cores': [1040, 'sub_int', 1, (1, 2, 3, 4, 5, 6, 7, 8),
                      'number of cores',
                      'Select amount of cores to be used for' +
                      ' PIV evaluations.'],

            'frequencing_sub_frame': [1045, 'sub_labelframe', None, None,
                                      'image frequencing', None],

            'sequence': [1050, 'sub', '(1+2),(3+4)',
                         ('(1+2),(2+3)', '(1+2),(3+4)'),
                         'sequence order',
                         'Select sequence order for evaluation.'],

            'skip': [1051, 'sub_int', 1, None, 'jump',
                     'Select sequence order jump for evaluation.' +
                     '\nEx: (1+(1+x)),(2+(2+x))'],

            'filters_sub_frame': [1100, 'sub_labelframe', None, None,
                                  'listbox filters', None],

            'load_settings':
                [1210, 'sub_bool', True, None,
                 'settings for using pandas',
                 'Individual settings ' +
                 'for loading files using pandas.'],

            'skiprows':
                [1211, 'sub', '0', None,
                 'skip rows',
                 'Number of rows skipped at the beginning of the file.'],

            'decimal':
                [1212, 'sub', '.', None,
                 'decimal separator',
                 'Decimal separator for floating point numbers.'],

            'sep':
                [1213, 'sub', 'tab', (',', ';', 'space', 'tab'),
                 'column separator',
                 'Column separator.'],

            'header':
                [1214, 'sub_bool', False, None,
                 'read header',
                 'Read header. ' +
                 'If chosen, first line will be interpreted as the header.' +
                 'Otherwise first line will be replaced with header names' +
                 'specified in the text field below.'],

            'header_names':
                [1215, 'sub', 'x,y,vx,vy,val,sig2noise', None,
                 'specify header names',
                 'Specify comma separated list of column names.' +
                 'Example: x,y,vx,vy,sig2noise'],

            'save_sub_frame':
                [1300, 'sub_labelframe', None,
                 None,
                 'PIV save settings',
                 None],

            'vec_fname':
                [1310, 'sub', 'vec', None,
                 'base output filename',
                 'Filename for vector output. A number and an acronym ' +
                 'that indicates the process history are added ' +
                 'automatically.'],

            'separator':
                [1320, 'sub', 'tab', (',', ';', 'space', 'tab'),
                 'delimiter',
                 'Delimiter.'],

            'image_plotting_sub_frame':
                [1500, 'sub_labelframe', None,
                 None,
                 'image plotting',
                 None],

            'matplot_intensity':
                [2032, 'sub_int', 255, None,
                 'reference intensity',
                 'Define a reference intensity for the plotting of images.'],

            # preprocessing
            'preproc':
                [2000, None, None, None,
                 'Preprocessing',
                 None],

            'preprocess_frame':
                [2005, 'labelframe', None,
                 None,
                 'Preprocessing',
                 None],

            'preprocess-info':
                [2010, 'label', None, None,

                 'All images are normalized to [0,1] float, \npreprocessed, ' +
                 'and resized to user defined value.',
                 None],

            'Invert_spacer':
                [2015, 'h-spacer', None,
                 None,
                 None,
                 None],

            'invert':
                [2020, 'bool', 'False', None,
                 'invert image',
                 'Invert image (see skimage invert()).'],

            'background_spacer':
                [2025, 'h-spacer', None,
                 None,
                 None,
                 None],

            'background_subtract':
                [2030, 'bool', 'False', None,
                 'subtract background',
                 'Subtract background via local sliding windows.'],

            'background_type':
                [2031, 'str', 'global mean', ('global mean', 'minA - minB'),
                 'background algorithm',
                 'The algorithm used to generate the background which is subtracted ' +
                 'from the piv images. ' +
                 'Warning: »minA - minB« is still in development, so it may not perform ' +
                 'to standard.'],

            'starting_frame':
                [2032, 'int', 0, None,
                 'starting image',
                 'Defining the starting image of the background subtraction.'],

            'ending_frame':
                [2033, 'int', 3, None,
                 'ending image',
                 'Defining the ending image of the background subtraction.'],

            'crop_ROI_spacer':
                [2035, 'h-spacer', None,
                 None,
                 None,
                 None],

            'crop_ROI':
                [2040, 'bool', 'False', None,
                 'crop region of interest',
                 'Crop region of interest. Allows images with different sizes to ' +
                 'have a uniform size after cropping.'],

            'crop_roi-xminmax':
                [2041, 'str', '200,800', None,
                 'x min/max',
                 "Define left/right side of region of interest by 'min,max'."],

            'crop_roi-yminmax':
                [2042, 'str', '200,800', None,
                 'y min/max',
                 "Define top/bottom of region of interest by 'min,max.'"],

            # 'dynamic_mask_spacer': # failed testing, needs fixing/testing
            #    [2045, 'h-spacer', None,
            #     None,
            #     None,
            #     None],

            # 'dynamic_mask':
            #    [2050, 'bool', 'False', None,
            #     'dynamic masking',
            #     'Dynamic masking for masking of images. \n' +
            #     'Warning: This is still in testing and is not recommended for use.'],

            # 'dynamic_mask_type':
            #    [2051, 'str', 'edge',
            #     ('edge', 'intensity'),
            #     'mask type',
            #     'Defining dynamic mask type.'],

            # 'dynamic_mask_threshold':
            #    [2052, 'float', 0.01, None,
            #     'mask threshold',
            #     'Defining threshold of dynamic mask.'],

            # 'dynamic_mask_size':
            #    [2053, 'int', 7, None,
            #     'mask filter size',
            #     'Defining size of the masks.'],
            'img_int_spacer': [2043, 'h-spacer', None, None, None, None],
            'img_int_resize': [2043, 'int', 255, None, 'resize intensity',
                               'Resize the image intensity to \n[0,x],'
                               ' where x is a user defined value.'],


            # processing
            'piv':
                [3000, None, None, None,
                 'PIV',
                 None],

            'piv_frame':
                [3005, 'labelframe', None,
                 None,
                 'Algorithms/Calibration',
                 None],

            # 'evaluation_method':
            #    [3010, 'string', 'FFT WinDef',
            #     ('Direct Correlation', 'FFT WinDef'),
            #     'evaluation method',
            #     'Direct Correlation: ' +
            #     'Direct correlation with extended size of the ' +
            #     'search area. \n' +
            #     'FFT WinDef: ' +
            #     'Fast Fourier Transforms with window deformation ' +
            #     '(recommended).'],

            'corr_method':
                [3020, 'str', 'circular',
                 ('circular', 'linear'),
                 'correlation method',
                 'Correlation method. Circular is no padding and' +
                 'linear is zero padding.'],

            'subpixel_method':
                [3030, 'str', 'gaussian',
                 ('centroid', 'gaussian', 'parabolic'),
                 'subpixel method',
                 'Fit function for determining the subpixel position ' +
                 'of the correlation peak.'],

            'sig2noise_method':
                [3040, 'string', 'peak2peak',
                 ('peak2peak', 'peak2mean'),
                 'signal2noise calc. method',
                 'Calculation method for the signal to noise ratio.'],

            's2n_mask':
                [3045, 'int', 2, None,
                 'signal to noise mask',
                 'the half size of the region around the first correlation'
                 ' peak to ignore for finding the second peak. Only used'
                 ' if sig2noise method = \'peak2peak\' '],

            'deformation_method':
                [3047, 'str', 'symmetric', ('symmetric', 'second image'),
                 'deformation method',
                 'Window deformation method. ' +
                 '»symmetric« deforms both first and second images. ' +
                 '\n»second image« deforms the second image only.'],

            'interpolation_order':
                [3048, 'int', 3, (0, 1, 2, 3, 4, 5),
                 'interpolation order',
                 'Interpolation oder of the spline window deformation. \n' +
                 '»0« yields zero order nearest interpolation \n' +
                 '»1« yields first order linear interpolation \n'
                 '»2« yields second order quadratic interpolation \n'
                 'and so on...'],

            'normalize_correlation':
                [3050, 'bool', False, None,
                 'normalize correlation',
                 'Normalize correlation.'],

            'calibration_spacer':
                [3055, 'h-spacer', None,
                 None,
                 None,
                 None],

            'dt':
                [3060, 'float', 1.0, None,
                 'dt',
                 'Interframing time in seconds.'],

            'scale':
                [3070, 'float', 1.0, None,
                 'scale',
                 'Interframing scaling in pix/m'],

            'flip_spacer':
                [3075, 'h-spacer', None,
                 None,
                 None,
                 None],

            'flip_u':
                [3080, 'bool', 'False', None,
                 'flip u-component',
                 'flip u-component array when saving RAW results.'],

            'flip_v':
                [3085, 'bool', 'False', None,
                 'flip v-component',
                 'flip v-component array when saving RAW results.'],

            'invert_spacer':
                [3090, 'h-spacer', None,
                 None,
                 None,
                 None],

            'invert_u':
                [3095, 'bool', 'False', None,
                 'invert u-component',
                 'Invert (negative) u-component when saving RAW results.'],

            'invert_v':
                [3096, 'bool', 'False', None,
                 'invert v-component',
                 'Invert (negative) v-component when saving RAW results.'],

            'swap_files_spacer':
                [3097, 'h-spacer', None,
                 None,
                 None,
                 None],

            'swap_files':
                [3098, 'bool', 'False', None,
                 'swap A/B files',
                 'Swap A/B files when analyzing.'],

            'windowing':
                [3100, None, None, None,
                 'Windowing',
                 None],

            'window_frame':
                [3105, 'labelframe', None,
                 None,
                 'Windowing',
                 None],

            # 'search_area':
            #    [3110, 'int', 64, (16, 32, 64, 128, 256),
            #     'search area',
            #     'Size of square search area in pixel for ' +
            #     'Single-pass DCC method.'],

            'corr_window':
                [3120, 'int', 32, (8, 16, 32, 64, 128),
                 'interrogation window',
                 'Size of the final interrogation windows in pixels.'],

            'overlap':
                [3130, 'int', 16, (4, 8, 16, 32, 64),
                 'overlap',
                 'Size of the final overlap in pixels.'],

            'coarse_factor':
                [3140, 'int', 3, (1, 2, 3, 4, 5),
                 'number of passes',
                 'Example: A window size of 16 and a number of refinement '
                 'steps of 3 gives an window size of 64×64 in the fist pass, '
                 '32×32 in the second pass and 16×16 pixel in the final '
                 'pass.'],

            'grid_refinement':
                [3150, 'str', 'all passes',
                 ('all passes', '2nd pass on', 'none'),
                 'grid refinement',
                 'Refine the interregationg grid every PIV pass when '
                 'performing multipass FFT. \n »all passes« refines all '
                 'passes. \n »2nd pass on« refines second pass on.'],

            'sub_window_frame':
                [3200, 'sub_labelframe', None,
                 None,
                 'custom windowing',
                 None],

            'custom_windowing':
                [3205, 'sub_bool', False, None,
                 'custom windowing',
                 'Enable custom windowing for more advanced techniques.'],

            'pass_1_spacer':
                [3210, 'sub_h-spacer', None,
                 None,
                 None,
                 None],

            'corr_window_1':
                [3220, 'sub_int', 256, None,
                 'interrogation window',
                 'Interrogation window for the first pass.'],

            'overlap_1':
                [3230, 'sub_int', 128, None,
                 'overlap',
                 'Size of the overlap of the first pass in pixels. The '
                 'overlap will then be calculated for the following passes.'],

            'pass_2_spacer':
                [3235, 'sub_h-spacer', None,
                 None,
                 None,
                 None],

            'pass_2':
                [3237, 'sub_bool', False, None,
                 'second pass',
                 'Enable a second pass in the FFT window deformation '
                 'evaluation.'],

            'corr_window_2':
                [3240, 'sub_int', 128, None,
                 'interrogation window',
                 'Interrogation window for the second pass.'],

            'pass_3_spacer':
                [3245, 'sub_h-spacer', None,
                 None,
                 None,
                 None],

            'pass_3':
                [3247, 'sub_bool', False, None,
                 'third pass',
                 'Enable a third pass in the FFT window deformation '
                 'evaluation.'],

            'corr_window_3':
                [3250, 'sub_int', 64, None,
                 'interrogation window',
                 'Interrogation window for the third pass.'],

            'pass_4_spacer':
                [3255, 'sub_h-spacer', None,
                 None,
                 None,
                 None],

            'pass_4':
                [3257, 'sub_bool', False, None,
                 'fourth pass',
                 'Enable a fourth pass in the FFT window deformation '
                 'evaluation.'],

            'corr_window_4':
                [3260, 'sub_int', 32, None,
                 'interrogation window',
                 'Interrogation window for the fourth pass.'],

            'pass_5_spacer':
                [3265, 'sub_h-spacer', None,
                 None,
                 None,
                 None],

            'pass_5':
                [3267, 'sub_bool', False, None,
                 'fifth pass',
                 'Enable a fifth pass in the FFT window deformation '
                 'evaluation.'],

            'corr_window_5':
                [3270, 'sub_int', 16, None,
                 'interrogation window',
                 'Interrogation window for the fifth pass.'],

            'pass_6_spacer':
                [3275, 'sub_h-spacer', None,
                 None,
                 None,
                 None],

            'pass_6':
                [3277, 'sub_bool', False, None,
                 'sixth pass',
                 'Enable a sixth pass in the FFT window deformation '
                 'evaluation.'],

            'corr_window_6':
                [3280, 'sub_int', 16, None,
                 'interrogation window',
                 'Interrogation window for the sixth pass.'],

            'pass_7_spacer':
                [3285, 'sub_h-spacer', None,
                 None,
                 None,
                 None],

            'pass_7':
                [3287, 'sub_bool', False, None,
                 'seventh pass',
                 'Enable a seventh pass in the FFT window deformation '
                 'evaluation.'],

            'corr_window_7':
                [3290, 'sub_int', 16, None,
                 'interrogation window',
                 'Interrogation window for the seventh pass.'],

            # individual pass validations
            'validation':
                [3300, None, None, None,
                 'Validation',
                 None],

            'validation_frame':
                [3305, 'labelframe', None,
                 None,
                 'Validation',
                 None],

            'piv_sub_frame1':
                [3306, 'sub_labelframe', None,
                 None,
                 'first pass validation',
                 None],

            'fp_local_med_threshold':
                [3310, 'sub_bool', True, None,
                 'local median validation',
                 'Discard vector, if the absolute difference with ' +
                 'the local median is greater than the threshold. '],

            'fp_local_med':
                [3311, 'sub_float', 1.2, None,
                 'local median threshold',
                 'Local median absolute difference threshold.'],

            'fp_local_med_size':
                [3312, 'sub_int', 1, None,
                 'local median kernel',
                 'Local median filter kernel size.'],

            'globa_thr_spacer':
                [3325, 'sub_h-spacer', None,
                 None,
                 None,
                 None],

            'fp_vld_global_threshold':
                [3330, 'sub_bool', False, None,
                 'global threshold validation',
                 'Validate first pass based on set global ' +
                 'thresholds.'],

            'fp_MinU':
                [3331, 'sub_float', -100.0, None,
                 'min u',
                 'Minimum U allowable component.'],

            'fp_MaxU':
                [3332, 'sub_float', 100.0, None,
                 'max u',
                 'Maximum U allowable component.'],

            'fp_MinV':
                [3333, 'sub_float', -100.0, None,
                 'min v',
                 'Minimum V allowable component.'],

            'fp_MaxV':
                [3334, 'sub_float', 100.0, None,
                 'max v',
                 'Maximum V allowable component.'],

            'piv_sub_frame2':
                [3340, 'sub_labelframe', None,
                 None,
                 'other pass validations',
                 None],

            'sp_local_med_validation':
                [3350, 'sub_bool', True, None,
                 'local median validation',
                 'Discard vector, if the absolute difference with ' +
                 'the local median is greater than the threshold.'],

            'sp_local_med':
                [3351, 'sub_float', 1.2, None,
                 'local median threshold',
                 'Local median absolute difference threshold.'],

            'sp_local_med_size':
                [3352, 'sub_int', 1, None,
                 'local median kernel',
                 'Local median filter kernel size.'],

            'glob_std_spacer':
                [3355, 'sub_h-spacer', None,
                 None,
                 None,
                 None],

            'sp_vld_std_threshold':
                [3360, 'sub_bool', False, None,
                 'standard deviation validation',
                 'Remove vectors, if the the sum of the squared ' +
                 'vector components is larger than the threshold ' +
                 'times the standard deviation of the flow field.'],

            'sp_std_threshold':
                [3361, 'sub_float', 8.0, None,
                 'std threshold',
                 'Standard deviation threshold.'],

            'glob_thr_spacer':
                [3365, 'sub_h-spacer', None,
                 None,
                 None,
                 None],

            'sp_vld_global_threshold':
                [3370, 'sub_bool', False, None,
                 'global threshold validation',
                 'Validate first pass based on set global ' +
                 'thresholds.'],

            'sp_vld_global_set_first':  # this needs some rewording
                [3371, 'sub_bool', True, None,
                 'set to first pass',
                 'Set the settings of the other pass validations ' +
                 'to the same as first pass.'],

            'sp_MinU':
                [3372, 'sub_float', -100.0, None,
                 'min u',
                 'Minimum U allowable component.'],

            'sp_MaxU':
                [3373, 'sub_float', 100.0, None,
                 'max u',
                 'Maximum U allowable component.'],

            'sp_MinV':
                [3374, 'sub_float', -100.0, None,
                 'min v',
                 'Minimum V allowable component.'],

            'sp_MaxV':
                [3375, 'sub_float', 100.0, None,
                 'max v',
                 'Maximum V allowable component.'],

            'individual_pass_postprocessing':
                [3380, None, None, None,
                 'PostProcessing',
                 None],

            'piv_pass_postprocessing_frame':
                [3383, 'labelframe', None,
                 None,
                 'Pass Postprocessing',
                 None],

            'piv_sub_frame3':
                [3385, 'sub_labelframe', None,
                 None,
                 'interpolation',
                 None],

            'adv_repl':
                [3390, 'sub_bool', True, None,
                 'replace vectors',
                 'Replace vectors between each pass.'],

            'adv_repl_method':
                [3391, 'sub', 'localmean',
                 ('localmean', 'disk', 'distance'),
                 'replacement method',
                 'Each NaN element is replaced by a weighed average' +
                 'of neighbours. Localmean uses a square kernel, ' +
                 'disk a uniform circular kernel, and distance a ' +
                 'kernel with a weight that is proportional to the ' +
                 'distance.'],

            'adv_repl_iter':
                [3392, 'sub_int', 10, None,
                 'number of iterations',
                 'If there are adjacent NaN elements, iterative ' +
                 'replacement is needed.'],

            'adv_repl_kernel':
                [3393, 'sub_int', 2, None,
                 'kernel size',
                 'Diameter of the NaN interpolation kernel.'],

            'piv_sub_frame4':
                [3395, 'sub_labelframe', None,
                 None,
                 'Smoothing',
                 None],

            'smoothn_each_pass':
                [3400, 'sub_bool', True, None,
                 'smoothen each pass',
                 'Smoothen each pass using openpiv.smoothn.'],

            'smoothn_first_more':
                [3401, 'sub_bool', False, None,
                 'double first pass strength',
                 'Double the smoothing strength on the first pass.'],

            'robust1':
                [3402, 'sub_bool', False, None,
                 'smoothen robust',
                 'Activate robust in smoothen (minimizes influence of '
                 'outlying data).'],

            'smoothn_val1':
                [3403, 'sub_float', 1.0, None,
                 'smoothing strength',
                 'Strength of smoothen script. Higher scalar number produces '
                 'more smoothed data.'],

            # validation/postprocessing
            'vld':
                [6000, None, None, None,
                 'PostProcessing',
                 None],

            'vld_frame':
                [6001, 'labelframe', None,
                 None,
                 'Postprocess',
                 None],

            'vld_sig2noise':
                [6010, 'bool', False, None,
                 'signal to noise ratio validation',
                 'Validate the data based on the signal to nose ratio ' +
                 'of the cross correlation.'],

            'sig2noise_threshold':
                [6030, 'float', 1.05, None,
                 's2n threshold',
                 'Threshold for filtering based on signal to noise ' +
                 'ratio. Recommended value: between 1.05 and 1.1.'],

            'horizontal_spacer11': [6035, 'h-spacer', None, None, None, None],

            'vld_global_std':
                [6040, 'bool', False, None,
                 'standard deviation validation',
                 'Validate the data based on a multiple of the ' +
                 'standard deviation.'],

            'global_std_threshold':
                [6050, 'float', 5.0, None,
                 'std threshold',
                 'Remove vectors, if the the sum of the squared ' +
                 'vector components is larger than the threshold ' +
                 'times the standard deviation of the flow field.'],

            'horizontal_spacer12':
                [6055, 'h-spacer', None,
                 None,
                 None,
                 None],

            'vld_local_med':
                [6060, 'bool', True, None,
                 'local median validation',
                 'Validate the data based on a local median ' +
                 'threshold.'],

            'local_median_threshold':
                [6065, 'float', 1.2, None,
                 'local median threshold',
                 'Discard vector, if the absolute difference with ' +
                 'the local median is greater than the threshold. '],

            'local_median_size':
                [6070, 'int', 1, None,
                 'local median kernel',
                 'Local median filter kernel size.'],

            'horizontal_spacer13':
                [6075, 'h-spacer', None,
                 None,
                 None,
                 None],

            'vld_global_thr':
                [6080, 'bool', False, None,
                 'global threshold validation',
                 'Validate the data based on set global ' +
                 'thresholds.'],

            'MinU':
                [6090, 'float', -30.0, None,
                 'min u',
                 'Minimum U allowable component.'],

            'MaxU':
                [6090, 'float', 30.0, None,
                 'max u',
                 'Maximum U allowable component.'],

            'MinV':
                [6090, 'float', -30.0, None,
                 'min v',
                 'Minimum V allowable component.'],

            'MaxV':
                [6090, 'float', 30.0, None,
                 'max v',
                 'Maximum V allowable component.'],

            'horizontal_spacer14':
                [6095, 'h-spacer', None,
                 None,
                 None,
                 None],

            'repl':
                [7010, 'bool', True, None,
                 'replace outliers',
                 'Replace outliers.'],

            'repl_method':
                [7020, 'str', 'localmean',
                 ('localmean', 'disk', 'distance'),
                 'replacement method',
                 'Each NaN element is replaced by a weighed average' +
                 'of neighbours. Localmean uses a square kernel, ' +
                 'disk a uniform circular kernel, and distance a ' +
                 'kernel with a weight that is proportional to the ' +
                 'distance.'],

            'repl_iter':
                [7030, 'int', 10, None,
                 'number of iterations',
                 'If there are adjacent NaN elements, iterative ' +
                 'replacement is needed.'],

            'repl_kernel':
                [7040, 'int', 2, None,
                 'kernel size',
                 'Diameter of the weighting kernel.'],

            'horizontal_spacer15':
                [7045, 'h-spacer', None,
                 None,
                 None,
                 None],

            'smoothn':
                [7050, 'bool', False, None,
                 'smoothn data',
                 'Smoothn data using openpiv.smoothn.'],

            'smoothn_type':
                [7060, 'str', 'last pass',
                 ('last pass', 'each pass'),
                 'smoothn vectors',
                 'Smoothn data with openpiv.smoothn. <each pass> only applies'
                 ' to windef'],

            'robust':
                [7070, 'bool', False, None,
                 'smoothn robust',
                 'Activate robust in smoothn (minimizes influence of outlying'
                 ' data).'],

            'smoothn_val':
                [7080, 'float', 1.0, None,
                 'smoothning strength',
                 'Strength of smoothn script. Higher scalar number produces ' +
                 'more smoothned data.'],

            'average_spacer':
                [7085, 'h-spacer', None,
                 None,
                 None,
                 None],

            'average_results':
                [7090, 'bool', False, None,
                 'average results (not implemented yet)',
                 'Average all results in selected directory. ' +
                 'Results in a single file with averaged results.'],

            'delimiter_spacer':
                [7095, 'h-spacer', None,
                 None,
                 None,
                 None],

            'delimiter':
                [7100, 'str', 'tab', (',', ';', 'space', 'tab'),
                 'delimiter',
                 'Delimiter.'],

            'postproc_spacer':
                [7105, 'h-spacer', None,
                 None,
                 None,
                 None],

            'post_button':
                [7110, 'post_button', None,
                 None,
                 None,
                 None],

            # plotting
            'plt':
                [8000, None, None, None,
                 'Plot',
                 None],
            'plt_frame':
                [8005, 'labelframe', None,
                 None,
                 'Plotting',
                 None],
            'plot_type':
                [8010, 'str', 'contour + vectors',
                 ('vectors', 'contour', 'contour + vectors',
                  'streamlines', 'histogram', 'profiles', 'scatter',
                  'line',
                  # 'bar', Failed testing (for Windows 10), simply locks GUI.
                  'density'),
                 'plot type',
                 'Select how to plot velocity data.'],
            'plot_title':
                [8020, 'str', 'Title', None,
                 'diagram title',
                 'diagram title.'],
            # plot_derivatives':
            #   [8075, 'str', 'None', ('None', 'Vorticity'),
            #   'plot derivatives',
            #   'Plot derivatives of the vector map (for vectors, countours,
            #   and streamlines only).'],
            'streamline_density':
                [8095, 'str', '0.5, 1', None,
                 'streamline density',
                 'streamline density. Can be one value (e.g. 1) or a couple' +
                 ' of values for a range (e.g. 0.5, 1).'],
            'integrate_dir':
                [8097, 'str', 'both', ('both', 'forward', 'backward'),
                 'streamline direction',
                 'Integrate the streamline in forward, backward or both ' +
                 'directions. default is both.'],
            'statistics_frame':
                [8105, 'labelframe', None,
                 None,
                 'Statistics',
                 None],
            'u_data':
                [8110, 'str', 'vx', None,
                 'x-data',
                 'Column name for the u-velocity component.' +
                 ' If unknown watch labbook entry.'],
            'v_data':
                [8120, 'str', 'vy', None,
                 'y-data',
                 'Column name for v-velocity component.' +
                 ' If unknown watch labbook entry.' +
                 ' For histogram only the v-velocity component is needed.'],
            'plot_scaling':
                [8130, 'str', 'None', ('None', 'logx', 'logy', 'loglog'),
                 'axis scaling', 'scales the axes. logarithm scaling x-axis' +
                 ' --> logx; logarithm scaling y-axis --> logy; ' +
                 'logarithm scaling both axes --> loglog.'],
            'histogram_type':
                [8140, 'str', 'bar',
                 ('bar', 'barstacked', 'step', 'stepfilled'),
                 'histogram type',
                 'Choose histogram type. Only available for histogram' +
                 'plot.'],
            'histogram_quantity':
                [8150, 'str', 'v_x', ('v', 'v_x', 'v_y'),
                 'histogram quantity',
                 'The absolute value of the velocity (v) or its x- ' +
                 'or y-component (v_x or v_y).'],
            'histogram_bins':
                [8160, 'int', 20, None,
                 'histogram number of bins',
                 'Number of bins (bars) in the histogram.'],
            'histogram_normalize':
                [8165, 'bool', False, None,
                 'normalize histogram',
                 'Normalize histogram (divide by the number of counts, '
                 'density).'],
            'profiles_orientation':
                [8170, 'str', 'vertical', ('vertical', 'horizontal'),
                 'profiles orientation',
                 'Plot v_y over x (horizontal) or v_x over y (vertical).'],
            'profiles_jump':
                [8180, 'int', 5, None,
                 'profile density',
                 'The amount of profile lines (minimum of 1).'],
            'plot_xlim':
                [8190, 'str', '', None,
                 'limits for the x-axis',
                 'For implementation use (lower_limit, upper_limit).'],
            'plot_ylim':
                [8200, 'str', '', None,
                 'limits for the y-axis',
                 'For implementation use (lower_limit, upper_limit).'],
            'modify_plot_appearance':
                [8500, None, None, None,
                 'Plot',
                 None],
            'modify_plot_frame':
                [8505, 'labelframe', None,
                 None,
                 'Modify Plot Appearance',
                 None],
            'vector_subframe':
                [8505, 'sub_labelframe', None,
                 None,
                 'Vectors',
                 None],
            'vec_scale':
                [8510, 'sub_int', 100, None,
                 'vector scaling',
                 'Velocity as a fraction of the plot width, e.g.: ' +
                 'm/s per plot width. Large values result in shorter ' +
                 'vectors.'],
            'vec_width':
                [8520, 'sub_float', 0.0025, None,
                 'vector line width',
                 'Line width as a fraction of the plot width.'],

            'invalid_color':
                [8530, 'dummy', 'red', None,
                 None, 'Choose the color of the vectors'],

            'valid_color':
                [8540, 'dummy', 'black', None,
                 None, 'Choose the color of the vectors'],

            # now applies to contours, so it is placed in the main labelframe
            'invert_yaxis':
                [8550, 'bool', True, None,
                 'invert y-axis',
                 'Define the top left corner as the origin ' +
                 'of the vector plot coordinate sytem, ' +
                 'as it is common practice in image processing.'],

            'derived_subframe':
                [8555, 'sub_labelframe', None,
                 None,
                 'Derived Parameters',
                 None],

            'color_map':
                [8560, 'sub', 'viridis', ('viridis', 'jet', 'short rainbow',
                                          'long rainbow', 'seismic', 'autumn',
                                          'binary'),
                 'Color map', 'Color map for streamline- and contour-plot.'],

            'extend_cbar':
                [8570, 'sub_bool', True, None,
                 'extend colorbar',
                 'Extend the top and bottom of the colorbar to accept out'
                 ' of range values.'],

            'velocity_color':
                [8575, 'sub', 'v', ('vx', 'vy', 'v'),
                 'set colorbar to: ',
                 'Set colorbar to velocity components.'],
            'color_levels':
                [8580, 'sub', '30', None, 'number of color levels',
                 'Select the number of color levels for contour plot.'],
            'vmin':
                [8590, 'sub', '', None,
                 'min velocity for colormap',
                 'minimum velocity for colormap (contour plot).'],
            'vmax':
                [8595, 'sub', '', None,
                 'max velocity for colormap',
                 'maximum velocity for colormap (contour plot).'],
            'statistics_subframe':
                [8600, 'sub_labelframe', None,
                 None,
                 'Statistics',
                 None],
            'plot_grid':
                [8610, 'sub_bool', True, None,
                 'grid',
                 'adds a grid to the diagram.'],
            'plot_legend':
                [8620, 'sub_bool', True, None,
                 'legend',
                 'adds a legend to the diagram.'],

            # lab-book
            'lab_book':
                [9000, None, None, None,
                 'Lab-Book',
                 None],

            'lab_book_content':
                [9010, 'text',
                 '',
                 None,
                 None,
                 None],

            'data_information':
                [9020, 'bool', False, None, 'log column information',
                 'shows column names, if you choose a file at the ' +
                 'right side.'],

            'used_addins':
                [99999, '[]', [], None, None, None]
        }

        # splitting the dictionary for more convenient access
        self.index = {}
        self.type = {}
        self.param = {}
        self.hint = {}
        self.label = {}
        self.help = {}
        self.add_parameters(self.default)

    def __getitem__(self, key):
        return self.param[key]

    def __setitem__(self, key, value):
        self.param[key] = value

    def load_settings(self, fname):
        """
            Read parameters from a JSON file.

            Args:
                fname (str): Path of the settings file in JSON format.

            Reads only parameter values. Content of the fields index,
            type, hint, label and help are always read from the default
            dictionary. The default dictionary may contain more entries
            than the JSON file (ensuring backwards compatibility).
        """
        try:
            f = open(fname, 'r')
            p = json.load(f)
            f.close()
        except BaseException:
            print('File not found: ' + fname)
        else:
            for key in self.param:
                if key in p:
                    self.param[key] = p[key]

    def dump_settings(self, fname):
        """
            Dump parameter values to a JSON file.

            Args:
                fname (str): A filename.

            Only the parameter values are saved. Other data like
            index, hint, label and help should only be defined in the
            default dictionary in this source code.
        """
        try:
            f = open(fname, 'w')
            json.dump(self.param, f)
            f.close()
        except BaseException:
            print('Unable to save settings: ' + fname)

    def generate_parameter_documentation(self, group=None):
        """
            Return parameter labels and help as reStructuredText def list.

            Parameters
            ----------
            group : int
                Parameter group.
                (e.g. OpenPivParams.PIVPROC)

            Returns
            -------
            str : A reStructuredText definition list for documentation.
        """
        s = ''
        for key in self.default:
            if (group < self.index[key] < group + 1000
                    and self.type[key] not in [
                        'labelframe',
                        'sub_labelframe',
                        'h-spacer',
                        'sub_h-spacer',
                        'dummy'
            ]):
                s = s + str(self.label[key]) + '\n' + "    " \
                    + str.replace(str(self.help[key]), '\n', '\n    ') + '\n\n'
        return s

    def add_parameters(self, param: dict) -> None:
        """
            splitting the dictionary for more convenient access
            outsourcing out of the init method was necessary to use it
            within the Add_In_Handler
            :param param: dictionary containing a list for each parameter
            :type param: dict
            :return: None
        """
        self.index.update(dict(zip(param.keys(),
                                   [val[0] for val in param.values()])))
        self.type.update(dict(zip(param.keys(),
                                  [val[1] for val in param.values()])))
        self.param.update(dict(zip(param.keys(),
                                   [val[2] for val in param.values()])))
        self.hint.update(dict(zip(param.keys(),
                                  [val[3] for val in param.values()])))
        self.label.update(dict(zip(param.keys(),
                                   [val[4] for val in param.values()])))
        self.help.update(dict(zip(param.keys(),
                                  [val[5] for val in param.values()])))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Parallel Processing of PIV images."""

from openpivgui.PreProcessing import gen_background, process_images
from openpivgui.open_piv_gui_tools import create_save_vec_fname, _round
import numpy as np
import time
import openpiv.smoothn as piv_smt
import openpiv.scaling as piv_scl
import openpiv.filters as piv_flt
import openpiv.validation as piv_vld
import openpiv.windef as piv_wdf
import openpiv.pyprocess as piv_prc
import openpiv.preprocess as piv_pre
import openpiv.tools as piv_tls
__licence__ = '''
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

__email__ = 'vennemann@fh-muenster.de'


class MultiProcessing(piv_tls.Multiprocesser):
    """
        Parallel processing, based on the corrresponding OpenPIV class.

        Do not run from the interactive shell or within IDLE! Details at:
        https://docs.python.org/3.6/library/multiprocessing.html#using-a-pool-of-workers

        Parameters
        ----------
        params : OpenPivParams
            A parameter object.
    """

    def __init__(self, gui):
        """
            Standard initialization method.

            For separating GUI and PIV code, the output filenames are
            generated here and not in OpenPivGui. In this way, this object
            might also be useful independently from OpenPivGui.
        """
        self.p = gui.p
        self.GUI = gui

        # generate background if needed
        if self.p['background_subtract']\
                and self.p['background_type'] != 'minA - minB':
            self.background = gen_background(self.p)
        else:
            self.background = None

        # custom image sequence with (1+[1+x]), (2+[2+x]) and ((1+[1+x]),
        # (3+[3+x]))
        if self.p['sequence'] == '(1+2),(2+3)':
            step = 1
        else:
            step = 2
        self.files_a = self.p['fnames'][0::step]
        self.files_b = self.p['fnames'][self.p['skip']::step]

        # making sure files_a is the same length as files_b
        diff = len(self.files_a) - len(self.files_b)
        if diff != 0:
            for i in range(diff):
                self.files_a.pop(len(self.files_b))
        print('Number of a files: ' + str(len(self.files_a)))
        print('Number of b files: ' + str(len(self.files_b)))

        if self.p['swap_files']:
            self.files_a, self.files_b = self.files_b, self.files_a

        self.n_files = len(self.files_a)
        self.save_fnames = []

        evaluation_method = 'FFT'

        postfix = '_piv_' + evaluation_method + '_'
        for n in range(self.n_files):
            self.save_fnames.append(
                create_save_vec_fname(path=self.files_a[n],
                                      basename=self.p['vec_fname'],
                                      postfix=postfix,
                                      count=n,
                                      max_count=self.n_files))

        # setup widgets so the user could still use the GUI when processing
        self.parameter = {}
        for key in self.p.param:
            key_type = self.p.type[key]
            if key_type not in ['labelframe',
                                'sub_labelframe',
                                'h-spacer',
                                'sub_h-spacer',
                                'dummy']:
                if 1030 < self.p.index[key] < 4000:
                    self.parameter[key] = self.p[key]

    def get_save_fnames(self):
        """
            Return a list of result filenames.

            Returns:
                str[]: List of filenames with resulting PIV data.
        """
        return self.save_fnames

    def get_num_frames(self) -> int:
        """
            Return the amount of image pairs that will be processed.

            Returns:
                int: The number of image pairs to be processed
        """
        return len(self.files_a)

    def process(self, args):
        """
            Process chain as configured in the GUI.

            Parameters
            ----------
            args : tuple
                Tuple as expected by the inherited run method:
                file_a (str) -- image file a
                file_b (str) -- image file b
                counter (int) -- index pointing to an element of the filename
                                 list
        """
        file_a, file_b, counter = args
        frame_a = piv_tls.imread(file_a)
        frame_b = piv_tls.imread(file_b)

        # Smoothning script borrowed from openpiv.windef
        s = self.p['smoothn_val']

        def smoothn(u, s):
            s = s
            u, _, _, _ = piv_smt.smoothn(
                u, s=s, isrobust=self.p['robust'])
            return(u)

        # delimiters placed here for safety
        delimiter = self.p['separator']
        if delimiter == 'tab':
            delimiter = '\t'
        if delimiter == 'space':
            delimiter = ' '

        # preprocessing
        print('\nPre-pocessing image pair: {}'.format(counter + 1))
        if self.p['background_subtract'] \
                and self.p['background_type'] == 'minA - minB':
            self.background = gen_background(self.p, frame_a, frame_b)

        frame_a = frame_a.astype(np.int32)
        frame_a = process_images(self, frame_a, self.GUI.preprocessing_methods,
                                 background=self.background)
        frame_b = frame_b.astype(np.int32)
        frame_b = process_images(self, frame_b, self.GUI.preprocessing_methods,
                                 background=self.background)

        print('Evaluating image pair: {}'.format(counter + 1))

        # evaluation first pass
        start = time.time()
        passes = 1
        # setup custom windowing if selected
        if self.parameter['custom_windowing']:
            corr_window_0 = self.parameter['corr_window_1']
            overlap_0 = self.parameter['overlap_1']
            for i in range(2, 8):
                if self.parameter['pass_%1d' % i]:
                    passes += 1
                else:
                    break

        else:
            passes = self.parameter['coarse_factor']
            if self.parameter['grid_refinement'] == 'all passes' \
                    and self.parameter['coarse_factor'] != 1:
                corr_window_0 = self.parameter['corr_window'] * \
                    2**(self.parameter['coarse_factor'] - 1)
                overlap_0 = self.parameter['overlap'] * \
                    2**(self.parameter['coarse_factor'] - 1)

            # Refine all passes after first when there are more than 1 pass.
            elif self.parameter['grid_refinement'] == '2nd pass on' \
                    and self.parameter['coarse_factor'] != 1:
                corr_window_0 = self.parameter['corr_window'] * \
                    2**(self.parameter['coarse_factor'] - 2)
                overlap_0 = self.parameter['overlap'] * \
                    2**(self.parameter['coarse_factor'] - 2)

            # If >>none<< is selected or something goes wrong, the window
            # size would remain the same.
            else:
                corr_window_0 = self.parameter['corr_window']
                overlap_0 = self.parameter['overlap']
        overlap_percent = overlap_0 / corr_window_0
        sizeX = corr_window_0

        u, v, sig2noise = piv_wdf.extended_search_area_piv(
            frame_a.astype(np.int32),
            frame_b.astype(np.int32),
            window_size=corr_window_0,
            overlap=overlap_0,
            search_area_size=corr_window_0,
            width=self.parameter['s2n_mask'],
            subpixel_method=self.parameter['subpixel_method'],
            sig2noise_method=self.parameter['sig2noise_method'],
            correlation_method=self.parameter['corr_method'],
            normalized_correlation=self.parameter['normalize_correlation'])

        x, y = piv_wdf.get_coordinates(frame_a.shape,
                                       corr_window_0,
                                       overlap_0)

        # validating first pass
        mask = np.full_like(x, 0)
        if self.parameter['fp_vld_global_threshold']:
            u, v, Mask = piv_vld.global_val(
                u, v,
                u_thresholds=(self.parameter['fp_MinU'],
                              self.parameter['fp_MaxU']),
                v_thresholds=(self.parameter['fp_MinV'],
                              self.parameter['fp_MaxV']))
            # consolidate effects of mask
            mask += Mask

        if self.parameter['fp_local_med']:
            u, v, Mask = piv_vld.local_median_val(
                u, v,
                u_threshold=self.parameter['fp_local_med'],
                v_threshold=self.parameter['fp_local_med'],
                size=self.parameter['fp_local_med_size'])
            mask += Mask

        if self.parameter['adv_repl']:
            u, v = piv_flt.replace_outliers(
                u, v,
                method=self.parameter['adv_repl_method'],
                max_iter=self.parameter['adv_repl_iter'],
                kernel_size=self.parameter['adv_repl_kernel'])
        print('Validated first pass result of image pair: {}.'
              .format(counter + 1))

        # smoothning  before deformation if 'each pass' is selected
        if self.parameter['smoothn_each_pass']:
            if self.parameter['smoothn_first_more']:
                s *= 2
            u = smoothn(u, s)
            v = smoothn(v, s)
            print('Smoothned pass 1 for image pair: {}.'.format(counter + 1))
            s = self.parameter['smoothn_val1']

        print('Finished pass 1 for image pair: {}.'.format(counter + 1))
        print("window size: " + str(corr_window_0))
        print('overlap: ' + str(overlap_0), '\n')

        # evaluation of all other passes
        if passes != 1:
            iterations = passes - 1
            for i in range(2, passes + 1):
                # setting up the windowing of each pass
                if self.parameter['custom_windowing']:
                    corr_window = self.parameter['corr_window_%1d' % i]
                    overlap = int(corr_window * overlap_percent)

                else:
                    if self.parameter['grid_refinement'] == 'all passes' or \
                            self.parameter['grid_refinement'] == '2nd pass on':
                        corr_window = self.parameter['corr_window'] * \
                            2**(iterations - 1)
                        overlap = self.parameter['overlap'] * \
                            2**(iterations - 1)

                    else:
                        corr_window = self.parameter['corr_window']
                        overlap = self.parameter['overlap']
                sizeX = corr_window

                # translate settings to windef settings object
                piv_wdf_settings = piv_wdf.Settings()
                piv_wdf_settings.correlation_method = \
                    self.parameter['corr_method']
                piv_wdf_settings.normalized_correlation = \
                    self.parameter['normalize_correlation']
                piv_wdf_settings.windowsizes = (corr_window,) * (passes + 1)
                piv_wdf_settings.overlap = (overlap,) * (passes + 1)
                piv_wdf_settings.num_iterations = passes
                piv_wdf_settings.subpixel_method = \
                    self.parameter['subpixel_method']
                piv_wdf_settings.deformation_method = \
                    self.parameter['deformation_method']
                piv_wdf_settings.interpolation_order = \
                    self.parameter['interpolation_order']
                piv_wdf_settings.sig2noise_validate = True,
                piv_wdf_settings.sig2noise_method = \
                    self.parameter['sig2noise_method']
                piv_wdf_settings.sig2noise_mask = self.parameter['s2n_mask']

                # do the correlation
                x, y, u, v, sig2noise, mask = piv_wdf.multipass_img_deform(
                    frame_a.astype(np.int32),
                    frame_b.astype(np.int32),
                    i,  # current iteration
                    x, y, u, v,
                    piv_wdf_settings)

                # validate other passes
                if self.parameter['sp_vld_global_threshold']:
                    u, v, Mask = piv_vld.global_val(
                        u, v,
                        u_thresholds=(self.parameter['sp_MinU'],
                                      self.parameter['sp_MaxU']),
                        v_thresholds=(self.parameter['sp_MinV'],
                                      self.parameter['sp_MaxV']))
                    mask += Mask  # consolidate effects of mask

                if self.parameter['sp_vld_global_threshold']:
                    u, v, Mask = piv_vld.global_std(
                        u, v,
                        std_threshold=self.parameter['sp_std_threshold'])
                    mask += Mask

                if self.parameter['sp_local_med_validation']:
                    u, v, Mask = piv_vld.local_median_val(
                        u, v,
                        u_threshold=self.parameter['sp_local_med'],
                        v_threshold=self.parameter['sp_local_med'],
                        size=self.parameter['sp_local_med_size'])
                    mask += Mask

                if self.parameter['adv_repl']:
                    u, v = piv_flt.replace_outliers(
                        u, v,
                        method=self.parameter['adv_repl_method'],
                        max_iter=self.parameter['adv_repl_iter'],
                        kernel_size=self.parameter['adv_repl_kernel'])
                print('Validated pass {} of image pair: {}.'
                      .format(i, counter + 1))

                # smoothning each individual pass if 'each pass' is selected
                if self.parameter['smoothn_each_pass']:
                    u = smoothn(u, s)
                    v = smoothn(v, s)
                    print('Smoothned pass {} for image pair: {}.'
                          .format(i, counter + 1))

                print('Finished pass {} for image pair: {}.'
                      .format(i, counter + 1))
                print("window size: " + str(corr_window))
                print('overlap: ' + str(overlap), '\n')
                iterations -= 1

        if self.p['flip_u']:
            u = np.flipud(u)

        if self.p['flip_v']:
            v = np.flipud(v)

        if self.p['invert_u']:
            u *= -1

        if self.p['invert_v']:
            v *= -1

        # scaling
        u = u / self.parameter['dt']
        v = v / self.parameter['dt']
        x, y, u, v = piv_scl.uniform(x, y, u, v,
                                     scaling_factor=self.parameter['scale'])
        end = time.time()

        # save data to file.
        out = np.vstack([m.ravel() for m in [x, y, u, v, mask, sig2noise]])
        np.savetxt(self.save_fnames[counter], out.T, fmt='%8.4f',
                   delimiter=delimiter)
        print('Processed image pair: {}'.format(counter + 1))

        sizeY = sizeX
        sizeX = ((int(frame_a.shape[0] - sizeX)
                  // (sizeX - (sizeX * overlap_percent))) + 1)
        sizeY = ((int(frame_a.shape[1] - sizeY)
                  // (sizeY - (sizeY * overlap_percent))) + 1)

        time_per_vec = _round((((end - start) * 1000)
                               / ((sizeX * sizeY) - 1)), 3)

        print('Process time: {} second(s)'.format((_round((end - start), 3))))
        print('Number of vectors: {}'.format(int((sizeX * sizeY) - 1)))
        print('Time per vector: {} millisecond(s)'.format(time_per_vec))



import tkinter as tk
import tkinter.ttk as ttk
import tkinter.messagebox as messagebox
import openpiv.tools as piv_tls

# A lot of optimization could be done in this file.


# check number of images, image types, and window sizing
def check_PIVprocessing(self):
    self.p = self
    '''Error checking'''
    # making sure there are 2 or more files loaded
    message = 'Please import two or more image files'
    if len(self.p['fnames']) < 2:
        if self.p['warnings']:
            messagebox.showwarning(title='Error Message',
                                   message=message)
        raise Exception(message)

    # checking for images
    message = "Please supply image files in 'bmp'," \
              " 'tiff', 'tif', 'TIF', 'jpg', 'jpeg', 'png', 'pgm'."
    test = self.p['fnames'][0]
    ext = test.split('.')[-1]
    if ext not in ['bmp', 'tiff', 'tif', 'TIF', 'jpg', 'jpeg', 'png', 'pgm']:
        if self.p['warnings']:
            messagebox.showwarning(title='Error Message',
                                   message=message)
        raise Exception(message)

    # checking interrogation window sizes in an inefficent manner (for now)
    test = piv_tls.imread(test)
    if 8 != 1:  # too lazy to fix spacing
        message = 'Please lower your starting interrogation window size.'
        if self.p['custom_windowing']:
            # making sure that the initial window is not too large
            if ((test.shape[0] / self.p['corr_window_1']) < 3 or
                    (test.shape[1] / self.p['corr_window_1']) < 3):
                if self.p['warnings']:
                    messagebox.showwarning(title='Error Message',
                                           message=message)
                raise ValueError(message)
            # making sure each pass has a decreasing interrogation window
            Message = 'Plase make sure that the custom windowing is' \
                      ' decresing with each pass.'
            window = self.p['corr_window_1']
            for i in range(2, 8):
                if self.p['pass_%1d' % i]:
                    if window >= self.p['corr_window_%1d' % i]:
                        window = self.p['corr_window_%1d' % i]

                    else:
                        if self.p['warnings']:
                            messagebox.showwarning(title='Error Message',
                                                   message=Message)
                        raise ValueError(Message)
                else:
                    break
        else:
            # checking the windowing for different multi pass setups
            message = 'Please lower your starting interrogation window' \
                      ' size or change multipass/grid refinement settings.'
            if self.p['grid_refinement'] == 'all passes' \
                    and self.p['coarse_factor'] != 1:
                if ((test.shape[0] /
                     (self.p['corr_window'] *
                      2**(self.p['coarse_factor'] -
                          1))) < 2.5 or (test.shape[1] /
                                         (self.p['corr_window'] *
                                          2**(self.p['coarse_factor'] -
                                              1))) < 2.5):
                    if self.p['warnings']:
                        messagebox.showwarning(title='Error Message',
                                               message=message)
                    raise ValueError(message)

            elif self.p['grid_refinement'] == '2nd pass on' \
                    and self.p['coarse_factor'] != 1:
                if ((test.shape[0] /
                     (self.p['corr_window'] *
                      2**(self.p['coarse_factor'] -
                          2))) < 2.5 or (test.shape[1] /
                                         (self.p['corr_window'] *
                                          2**(self.p['coarse_factor'] -
                                              2))) < 2.5):
                    if self.p['warnings']:
                        messagebox.showwarning(title='Error Message',
                                               message=message)
                    raise ValueError(message)

            else:
                if ((test.shape[0] / self.p['corr_window']) < 3 or
                        (test.shape[1] / self.p['corr_window']) < 3):
                    if self.p['warnings']:
                        messagebox.showwarning(title='Error Message',
                                               message=message)
                    raise ValueError(message)


def check_processing(self):  # check for threads
    self = self
    message = 'Please stop all threads/processes to start processing.'
    checker = 0
    # check if any threads are alive
    try:
        if self.processing_thread.is_alive():
            if self.p['warnings']:
                messagebox.showwarning(title='Error Message',
                                       message=message)
            checker += 1
    except BaseException:
        pass

    try:
        if self.postprocessing_thread.is_alive():
            if self.p['warnings']:
                messagebox.showwarning(title='Error Message',
                                       message=message)
            checker += 1
    except BaseException:
        pass
    # if a thread is alive, an error shall be raised
    if checker != 0:
        # raising errors did not work in try statement for some reason
        raise Exception(message)


def check_postprocessing(self):  # check for file types
    self.p = self
    test = self.p['fnames'][0]  # testing if .vec or .txt files are loaded.
    ext = test.split('.')[-1]
    if ext not in ('txt', 'vec'):
        message = 'Please provide ASCI-II .vec or .txt files.'
        if self.p['warnings']:
            messagebox.showwarning(title='Error Message',
                                   message=message)
        raise Exception(message)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""A simple GUI for OpenPIV."""

import openpivgui.vec_plot as vec_plot
from openpivgui.open_piv_gui_tools import str2list, str2dict, get_dim, _round
from openpivgui.ErrorChecker import check_PIVprocessing, check_processing, \
    check_postprocessing
from openpivgui.PostProcessing import PostProcessing
from openpivgui.PreProcessing import gen_background, process_images
from openpivgui.MultiProcessing import MultiProcessing
from openpivgui.CreateToolTip import CreateToolTip
from openpivgui.OpenPivParams import OpenPivParams
from scipy.ndimage.filters import gaussian_filter, gaussian_laplace
from matplotlib.figure import Figure as Fig
from matplotlib.backend_bases import key_press_handler
from matplotlib.backends.backend_tkagg import (
    FigureCanvasTkAgg,
    NavigationToolbar2Tk)
import matplotlib.pyplot as plt
import openpiv.tools as piv_tls
import pandas as pd
import numpy as np
from tkinter import colorchooser
from datetime import datetime
import threading
import shutil
import webbrowser
import tkinter.messagebox as messagebox
import tkinter.ttk as ttk
import tkinter.filedialog as filedialog
import tkinter as tk
import inspect
import json
import sys
import re
import os
import AddInHandler

__version__ = '0.4.5'

__licence__ = '''
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

__email__ = 'vennemann@fh-muenster.de'


class ToggledFrame(tk.Frame):

    def __init__(self, parent, text="", *args, **options):
        tk.Frame.__init__(self, parent, *args, **options)

        self.show = tk.IntVar()
        self.show.set(0)

        self.title_frame = ttk.Frame(self)
        self.title_frame.pack(fill="x", expand=1)

        ttk.Label(self.title_frame, text=text).pack(side="left", fill="x",
                                                    expand=1)

        self.toggle_button = ttk.Checkbutton(self.title_frame, width=2,
                                             text='+', command=self.toggle,
                                             variable=self.show,
                                             style='Toolbutton')
        self.toggle_button.pack(side="left")

        self.sub_frame = tk.Frame(self, relief="sunken", borderwidth=1)

    def toggle(self):
        if bool(self.show.get()):
            self.sub_frame.pack(fill="x", expand=1)
            self.toggle_button.configure(text='-')
        else:
            self.sub_frame.forget()
            self.toggle_button.configure(text='+')


class OpenPivGui(tk.Tk):
    """
        Simple OpenPIV GUI

        Usage:

        1. Press »File« and then »Import files« or »Import directory«.
           Either select some image pairs (Ctrl + Shift) or a directory
           that contains image files.

        2. Click on the links in the file-list on the right to inspect
           the images.

        3. Walk through the drop-down-menues »General«, »Preprocessing«,
           and »Analysis« and edit the parameters.

        4. Press the »start processing« butten (bottom left), to
           start the processing chain.

        5. Inspect the results by clicking on the links in the file-list.
           Use the »Plot« drop-down menu for changing the plot parameters.

        6. Use the »back« and »forward« buttons to go back to the images,
           in case you want to repeat the evaluation.

        7. For post-processing, use the »back« and »forward« buttons«
           to list the vector files. Modify the Post-Processing
           parameters and hit the »start post-processing« button.

        See also:

        https://github.com/OpenPIV/openpiv_tk_gui
    """

    buttons = {}
    preprocessing_methods = {}
    postprocessing_methods = {}

    def __init__(self):
        """
            Standard initialization method.
        """
        print('Initializing GUI')
        self.VERSION = __version__
        self.TITLE = 'Simple OpenPIV GUI'
        tk.Tk.__init__(self)
        # path of gui folder
        self.path = os.path.dirname(os.path.abspath(__file__))
        # path for image or icon
        self.icon_path = os.path.join(self.path, 'res/icon.png')
        # convert .png into a usable icon photo
        self.iconphoto(False, tk.PhotoImage(file=self.icon_path))
        self.title(self.TITLE + ' ' + self.VERSION)
        # handle for user closing GUI through window manager
        self.protocol("WM_DELETE_WINDOW", self.destroy)
        # the parameter object
        self.p = OpenPivParams()
        self.p.load_settings(self.p.params_fname)
        # background variable for widget data:
        self.tkvars = {}
        # handle for settings frames on riders
        self.set_frame = []
        # handle for text-area objects
        self.ta = {}
        # handle for list-box
        self.lb = None
        print('Initializing widgets')
        # Initializing the addin parameters and methods
        Add_In_Handler.init_add_ins(self)
        self.__init_widgets()
        self.set_settings()
        self.log(timestamp=True, text='--------------------------------' +
                                      '\nTkinter OpenPIV session started.')
        self.log(text='OpenPivGui version: ' + self.VERSION)
        print('Initialized GUI, ready for processing')

    def start_processing(self):
        """Wrapper function to start processing in a separate thread."""
        try:
            self.get_settings()
            check_processing(self)  # simple error checking.
            check_PIVprocessing(self.p)
            self.processing_thread = threading.Thread(target=self.processing)
            self.processing_thread.start()
        except Exception as e:
            print('PIV evaluation thread stopped. ' + str(e))

    def processing(self):
        try:
            self.log(timestamp=True,
                     text='-----------------------------' +
                     '\nPre processing finished.',
                     group=self.p.PREPROC)
            '''Start the processing chain.

            This is the place to implement additional function calls.
            '''

            # parallel PIV evaluation:
            print('Starting evaluation.')
            self.progressbar.start()

            self.get_settings()
            mp = MultiProcessing(self)

            number_of_frames = mp.get_num_frames()
            self.process_type.config(text='Processing {} PIV image pair(s)'
                                     .format(number_of_frames))

            return_fnames = mp.get_save_fnames()

            # keep number of cores in check
            # if there are no cored available, then raise exception
            if os.cpu_count() == 0:
                raise Exception('Warning: no available threads to process in.')
            # allow for automatic or manual core selection
            if self.p['manual_select_cores']:
                cpu_count = self.p['cores']

            else:
                cpu_count = os.cpu_count()

            if "idlelib" in sys.modules:
                self.log('Running as a child of IDLE: '
                         'Deactivated multiprocessing.')
                cpu_count = 1

            if cpu_count >= os.cpu_count():
                raise Exception('Please lower the amount of cores ' +
                                'or deselect >manually select cores<.')

            print('Cores left: {} of {}.'.format(
                (os.cpu_count() - cpu_count), os.cpu_count()))

            mp.run(func=mp.process, n_cpus=cpu_count)

            # update file list with result vector files:
            self.tkvars['fnames'].set(return_fnames)
            self.log(timestamp=True,
                     text='\nPIV evaluation finished.',
                     group=self.p.PIVPROC)

            self.progressbar.stop()
            self.process_type.config(text='Processed {} PIV image pair(s)'
                                     .format(number_of_frames))

            # update file count
            self.get_settings()
            self.num_label.config(text=len(self.p['fnames']))
        except Exception as e:
            print('PIV evaluation thread stopped. ' + str(e))
            self.progressbar.stop()
            self.process_type.config(text='Failed to process image pair(s)')

    def start_postprocessing(self):
        """Wrapper function to start processing in a separate thread."""
        try:
            # if os.cpu_count() == 0:  # if there are no cored available,
            # then raise exception
            #    raise Exception('Warning: no available threads to
            #    process in.')
            check_processing(self)
            check_postprocessing(self.p)  # simple error checking
            self.postprocessing_thread = threading.Thread(
                target=self.postprocessing)
            self.postprocessing_thread.start()
        except Exception as e:
            print('Post-processing thread stopped. ' + str(e))

    def postprocessing(self):
        try:
            self.progressbar.start()
            self.process_type.config(text='Processing {} PIV result(s)'
                                     .format(len(self.p['fnames'])))

            print('Starting validation. Please wait for validation to finish')

            # used to include the validation methods of addins loaded
            boolean_vars_of_add_ins = []
            for func in self.postprocessing_methods:
                # check rather func is an validation or postprocessing addin
                if self.postprocessing_methods[func][0] == 'validation':
                    self.get_settings()
                    boolean_var = self.postprocessing_methods[func][1]
                    boolean_vars_of_add_ins.append(self.p[str(boolean_var)])
                    # check for checkbox which variable name is stored in
                    # boolean_var
                    if self.p[str(boolean_var)]:
                        post_proc = PostProcessing(self.p)
                        self.tkvars['fnames'].set(
                            self.postprocessing_methods[func][2]
                            (self, post_proc.delimiter))

            # standard deviation validation
            self.get_settings()
            if self.p['vld_sig2noise']:
                self.tkvars['fnames'].set(
                    PostProcessing(self.p).sig2noise())

            # standard deviation validation
            self.get_settings()
            if self.p['vld_global_std']:
                self.tkvars['fnames'].set(
                    PostProcessing(self.p).global_std())

            # global threshold validation
            self.get_settings()
            if self.p['vld_global_thr']:
                self.tkvars['fnames'].set(
                    PostProcessing(self.p).global_val())

            # local median validation
            self.get_settings()
            if self.p['vld_local_med']:
                self.tkvars['fnames'].set(
                    PostProcessing(self.p).local_median())

            # log validation parameters
            if (True in boolean_vars_of_add_ins or
                self.p['vld_sig2noise'] or
                self.p['vld_global_std'] or
                self.p['vld_global_thr'] or
                    self.p['vld_local_med']):
                self.log(timestamp=True,
                         text='\nValidation finished.',
                         group=self.p.VALIDATION)
            print('Finished validation. Please wait for postprocessing '
                  'to finish.')

            # used to include the postprocess methods of addins loaded
            boolean_vars_of_add_ins = []
            for func in self.postprocessing_methods:
                # check rather func is an validation or postprocessing addin
                if self.postprocessing_methods[func][0] == 'postprocessing':
                    self.get_settings()
                    boolean_var = self.postprocessing_methods[func][1]
                    boolean_vars_of_add_ins.append(self.p[str(boolean_var)])
                    # check for checkbox which variable name is stored in
                    # boolean_var
                    if self.p[str(boolean_var)]:
                        post_proc = PostProcessing(self.p)
                        self.tkvars['fnames'].set(
                            self.postprocessing_methods[func][2]
                            (self, post_proc.delimiter))

            # post processing
            self.get_settings()
            if self.p['repl']:
                self.tkvars['fnames'].set(
                    PostProcessing(self.p).repl_outliers())

            # smooth post processing
            self.get_settings()
            if self.p['smoothn']:
                self.tkvars['fnames'].set(
                    PostProcessing(self.p).smoothn_r())

            # average all ressults
            # self.get_settings()
            # if self.p['average_results']:
            #    self.tkvars['fnames'].set(
            #        PostProcessing(self.p).average())

            # log parameters
            if (True in boolean_vars_of_add_ins or
                self.p['repl'] or
                self.p['smoothn'] or
                    self.p['average_results']):
                self.log(timestamp=True,
                         text='\nPost processing finished.',
                         group=self.p.POSTPROC)
            print('Finished postprocessing.')

            self.progressbar.stop()
            self.process_type.config(text='Processed {} PIV result(s)'
                                     .format(len(self.p['fnames'])))

            # update file count
            self.get_settings()
            self.num_label.config(text=len(self.p['fnames']))
        except Exception as e:
            print('Postprocessing thread stopped. ' + str(e))
            self.progressbar.stop()
            self.process_type.config(text='Failed to postprocess results(s)')

    def __init_widgets(self):
        """Creates a widget for each variable in a parameter object."""
        self.__init_buttons()
        f = ttk.Frame(self)
        f.pack(side='left',
               fill='both',
               expand='True')
        # holds riders for parameters
        self.__init_notebook(f)
        # plotting area
        self.__init_fig_canvas(f)
        # variable widgets:
        for key in sorted(self.p.index, key=self.p.index.get):
            if self.p.type[key] == 'dummy':
                pass
            elif self.p.type[key] == 'bool':
                self.__init_checkbutton(key)
            elif self.p.type[key] == 'str[]':
                self.__init_listbox(key)
            elif self.p.type[key] == 'text':
                self.__init_text_area(key)
            elif self.p.type[key] == 'labelframe':
                self.__init_labelframe(key)
            elif self.p.type[key] == 'label':
                self.__init_label(key)
            elif self.p.type[key] == 'post_button':
                self.__init_post_button(key)
            elif self.p.type[key] == 'h-spacer':
                self.__init_horizontal_spacer(key)
            elif self.p.type[key] == 'sub_bool':
                self.__init_sub_checkbutton(key)
            elif self.p.type[key] == 'sub_labelframe':
                self.__init_sub_labelframe(key)
            elif self.p.type[key] == 'sub_h-spacer':
                self.__init_sub_horizontal_spacer(key)
            elif self.p.type[key] is None:
                self.__add_tab(key)
            else:
                self.__init_entry(key)

            # create widgets that are not in OpenPivParams
            if self.p.index[key] == 3500:
                self.__init_analysisframe(key)
            if self.p.index[key] == 8120:
                self.__init_vec_colorpicker(key)

    def __init_fig_canvas(self, mother_frame):
        """
            Creates a plotting area for matplotlib.

            Parameters
            ----------
            mother_frame : ttk.Frame
                A frame to place the canvas in.
        """
        self.fig = Fig()  # Enlargended canvas to my preference
        self.fig_frame = ttk.Frame(mother_frame)
        self.fig_frame.pack(side='left',
                            fill='both',
                            expand='True')

        self.fig_canvas = FigureCanvasTkAgg(
            self.fig, master=self.fig_frame)
        self.fig_canvas.draw()

        self.fig_canvas.get_tk_widget().pack(
            side='left',
            fill='x',
            expand='True')
        fig_toolbar = NavigationToolbar2Tk(self.fig_canvas,
                                           self.fig_frame)
        fig_toolbar.update()

        ttk.Button(self.fig_frame,
                   text='start processing',
                   command=self.start_processing).pack(side='left')
        ttk.Button(self.fig_frame,
                   text='start postprocessing',
                   command=self.start_postprocessing).pack(side='left')
        ttk.Button(self.fig_frame,
                   text='% invalid vectors',
                   command=self.calculate_invalid_vectors).pack(side='left')

        self.progressbar = ttk.Progressbar(self.fig_frame, orient='horizontal',
                                           length=200, mode='indeterminate')
        self.progressbar.pack(side='right')

        self.process_type = ttk.Label(self.fig_frame, text=' ')
        self.process_type.pack(side='right')

        self.fig_canvas._tkcanvas.pack(side='top',
                                       fill='both',
                                       expand='True')
        self.fig_canvas.mpl_connect("key_press_event",
                                    lambda: key_press_handler(event,
                                                              self.fig_canvas,
                                                              fig_toolbar))

    def __fig_toolbar_key_pressed(self, event):
        """Handles matplotlib toolbar events."""
        key_press_handler(event,
                          self.fig_canvas,
                          self.fig_toolbar)

    def __init_notebook(self, mother_frame):
        """The notebook is the root widget for tabs or riders."""
        style = ttk.Style()
        style.layout('TNotebook.Tab', [])
        self.nb = ttk.Notebook(mother_frame, width=260)
        self.nb.pack(side='right', fill='both', expand='False')

    def __add_tab(self, key):
        """Add an additional rider to the notebook."""
        self.set_frame.append(ttk.Frame(self.nb))
        self.nb.add(self.set_frame[-1], text=self.p.label[key])

    def __init_buttons(self):
        """Add buttons and bind them to methods."""
        f = ttk.Frame(self)
        files = ttk.Menubutton(f, text='File')
        options = tk.Menu(files, tearoff=0)
        files.config(menu=options)
        options.add_command(label='Import files',
                            command=self.select_image_files)
        options.add_command(label='Import directory',
                            command=self.open_directory)
        options.add_separator()
        options.add_command(label='Save session',
                            command=lambda: self.p.dump_settings(
                                filedialog.asksaveasfilename()))
        options.add_command(label='Load session', command=self.load_settings)
        options.add_command(label='Reset session', command=self.reset_params)
        options.add_separator()
        options.add_command(label='Move files', command=self.move_files)
        options.add_command(label='Delete files', command=self.delete_files)
        options.add_separator()
        options.add_command(label='Select Add Ins',
                            command=lambda:
                            Add_In_Handler.select_add_ins(self))
        options.add_command(label='Exit', command=self.destroy)
        files.pack(side='left', fill='x')

        general = ttk.Menubutton(f, text='General')
        options1 = tk.Menu(general, tearoff=0)
        general.config(menu=options1)
        options1.add_command(label='General settings',
                             command=lambda: self.selection(0))
        general.pack(side='left', fill='x')

        preproc = ttk.Menubutton(f, text='Preprocessing')
        options2 = tk.Menu(preproc, tearoff=0)
        preproc.config(menu=options2)
        options2.add_command(label='Preprocessing',
                             command=lambda: self.selection(1))
        preproc.pack(side='left', fill='x')

        piv = ttk.Menubutton(f, text='Analysis')
        options2 = tk.Menu(piv, tearoff=0)
        piv.config(menu=options2)
        options2.add_command(label='Algorithms\Calibration',
                             command=lambda: self.selection(2))
        options2.add_command(label='Windowing',
                             command=lambda: self.selection(3))
        options2.add_command(label='Validation',
                             command=lambda: self.selection(4))
        options2.add_command(label='Pass Postprocessing',
                             command=lambda: self.selection(5))
        options2.add_command(label='Start Analysis',
                             command=self.start_processing)
        piv.pack(side='left', fill='x')

        postproc = ttk.Menubutton(f, text='Postprocess')
        options3 = tk.Menu(postproc, tearoff=0)
        postproc.config(menu=options3)
        options3.add_command(label='Postprocess',
                             command=lambda: self.selection(6))
        options3.add_command(label='Start Postprocessing',
                             command=self.start_postprocessing)
        postproc.pack(side='left', fill='x')

        plot = ttk.Menubutton(f, text='Plotting')
        options4 = tk.Menu(plot, tearoff=0)
        plot.config(menu=options4)
        options4.add_command(
            label='Plotting', command=lambda: self.selection(7))
        options4.add_command(
            label='Modify Appearance', command=lambda: self.selection(8))
        plot.pack(side='left', fill='x')
        lab_func = ttk.Menubutton(f, text='Lab Book')
        options6 = tk.Menu(lab_func, tearoff=0)
        lab_func.config(menu=options6)
        options6.add_command(label='Show Lab Book',
                             command=lambda: self.selection(9))
        lab_func.pack(side='left', fill='x')

        usage_func = ttk.Menubutton(f, text='Usage')
        options7 = tk.Menu(usage_func, tearoff=0)
        usage_func.config(menu=options7)
        options7.add_command(label='Usage',
                             command=lambda: messagebox.showinfo(
                                 title='Help',
                                 message=inspect.cleandoc(
                                     OpenPivGui.__doc__)))
        usage_func.pack(side='left', fill='x')

        web_func = ttk.Menubutton(f, text='Web')
        options8 = tk.Menu(web_func, tearoff=0)
        web_func.config(menu=options8)
        options8.add_command(label='Web', command=self.readme)
        web_func.pack(side='left', fill='x')
        addin_func = ttk.Menubutton(f, text='Add_In Functions')
        addin_func_options = tk.Menu(addin_func, tearoff=0)
        addin_func.config(menu=addin_func_options)
        for button in self.buttons:
            self.buttons[button](self, addin_func_options)
        addin_func.pack(side='left', fill='x')
        f.pack(side='top', fill='x')

    def selection(self, num):
        self.nb.select(num)

    def calculate_invalid_vectors(self):
        try:
            self.get_settings()

            data = self.load_pandas(self.p['fnames'][self.index])
            data = data.to_numpy().astype(np.float)

            try:
                invalid = data[:, 4].astype('bool')
            except BaseException:
                invalid = np.asarray([True for i in range(len(data))])
                print('No typevectors found')

            invalid = np.count_nonzero(invalid)
            percent = _round(((invalid / len(data[:, 0])) * 100), 4)
            message = ('Percent invalid vectors for result index {}: {}%'
                       .format(self.index, percent))

            if self.p['pop_up_info']:
                messagebox.showinfo(title='Statistics',
                                    message=message)
            print(message)

        except Exception as e:
            print('Could not read file for calculating percent of'
                  ' invalid vectors.')
            print('Reason: ' + str(e))

    def reset_params(self):
        """Reset parameters to default values."""
        answer = messagebox.askyesno(
            title='Reset session',
            message='Reset all parameters to default values?')
        if answer:
            self.p = OpenPivParams()
            self.set_settings()

    def readme(self):
        """Opens https://github.com/OpenPIV/openpiv_tk_gui."""
        webbrowser.open('https://github.com/OpenPIV/openpiv_tk_gui')

    def delete_files(self):
        """Delete files currently listed in the file list."""
        answer = messagebox.askyesno(
            title='Delete files',
            message='Are you sure you want to delete selected files?')
        if answer:
            files = self.p['fnames'][:]
            for f in files:
                os.remove(f)
            self.navigate('back')

    def move_files(self):
        """Move files to a new place."""
        files = self.p['fnames'][:]
        dir = filedialog.askdirectory(mustexist=False)
        if len(dir) > 0:
            if not os.path.exists(dir):
                os.mkdir(dir)
            for src in files:
                dst = dir + os.path.sep + os.path.basename(src)
                shutil.move(src, dst)
            self.navigate('back')

    def load_settings(self):
        """Load settings from a JSON file."""
        settings = filedialog.askopenfilename()
        if len(settings) > 0:
            self.p.load_settings(settings)
            self.set_settings()

    def load_pandas(self, fname):
        """
            Load files in a pandas data frame.

            On the rider named General, the parameters for loading
            the data frames can be specified.
            No parameters have to be set for image processing.

            Parameters
            ----------
            fname :
                A filename.

            Returns
            -------
            pandas.DataFrame :
                In case of an error, the errormessage is returned (str).
        """
        sep = self.p['sep']
        if sep == 'tab':
            sep = '\t'
        if sep == 'space':
            sep = ' '

        ext = fname.split('.')[-1]
        if ext in ['txt', 'dat', 'jvc', 'vec', 'csv']:
            if self.p['load_settings']:
                if self.p['header']:
                    data = pd.read_csv(fname,
                                       decimal=self.p['decimal'],
                                       skiprows=int(self.p['skiprows']),
                                       sep=sep)
                elif not self.p['header']:
                    data = pd.read_csv(fname,
                                       decimal=self.p['decimal'],
                                       skiprows=int(self.p['skiprows']),
                                       sep=sep,
                                       header=0,
                                       names=self.p['header_names'].split(','))
            else:
                data = pd.read_csv(fname,
                                   decimal=',',
                                   skiprows=0,
                                   sep='\t',
                                   names=['x', 'y', 'vx', 'vy', 'sig2noise'])
        else:
            data = 'File could not be read. Possibly it is an image file.'
        return data

    def __init_listbox(self, key):
        """
            Creates an interactive list of filenames.

            Parameters
            ----------
            key : str
                Key of a settings object.
        """
        # root widget
        f = ttk.Frame(self)
        f.pack(side='bottom',
               fill='both',
               expand='True')
        # filter hint
        hint_frame = ttk.Frame(f)
        hint_title = ttk.Label(hint_frame, text=' filter: ')
        self.filter_hint = ttk.Label(hint_frame,
                                     text='None')
        hint_title.pack(anchor='nw', side='left')
        self.filter_hint.pack(anchor='nw')
        hint_frame.pack(side='top', fill='x', expand='False')

        # number of files
        num_frame = ttk.Frame(f)
        num_label = ttk.Label(num_frame, text=' number of files: ')
        self.num_label = ttk.Label(num_frame,
                                   text=len(self.p['fnames']))
        num_label.pack(anchor='nw', side='left')
        self.num_label.pack(anchor='nw')
        num_frame.pack(side='top', fill='x', expand='False')

        # scrolling
        sbx = ttk.Scrollbar(f, orient="horizontal")
        sbx.pack(side='top', fill='x')
        sby = ttk.Scrollbar(f, orient="vertical")
        sby.pack(side='right', fill='y')
        self.lb = tk.Listbox(f, yscrollcommand=sbx.set)
        self.lb = tk.Listbox(f, yscrollcommand=sby.set)
        sbx.config(command=self.lb.xview)
        sby.config(command=self.lb.yview)
        self.lb['width'] = 25

        # background variable
        self.tkvars.update({key: tk.StringVar()})
        self.tkvars[key].set(self.p['fnames'])
        self.lb['listvariable'] = self.tkvars[key]

        # interaction
        self.lb.bind('<<ListboxSelect>>', self.__listbox_selection_changed)
        self.lb.pack(side='top', fill='y', expand='True')

        # navigation buttons
        f = ttk.Frame(f)
        ttk.Button(f,
                   text='< back',
                   command=lambda: self.navigate('back')).pack(
            side='left', fill='x')
        ttk.Button(f,
                   text='forward >',
                   command=lambda: self.navigate('forward')).pack(
            side='right', fill='x')
        f.pack()

    def get_filelistbox(self):
        """
            Return a handle to the file list widget.

            Returns
            -------
            tkinter.Listbox
                A handle to the listbox widget holding the filenames
        """
        return self.lb

    def navigate(self, direction):
        '''Navigate through processing steps.

        Display a filtered list of files of the current
        directory. This function cycles through the filters
        specified by the key 'navi_pattern' in the settings object.

        Parameters
        ----------
        direction : str
            'back' or 'forward'.
        '''
        pattern_lst = str2list(self.p['navi_pattern'])
        dirname = os.path.dirname(self.p['fnames'][0])
        files = os.listdir(dirname)
        if direction == 'back':
            self.p.navi_position -= 1
            if self.p.navi_position == -1:
                self.p.navi_position = len(pattern_lst) - 1
        elif direction == 'forward':
            self.p.navi_position += 1
            if self.p.navi_position == len(pattern_lst):
                self.p.navi_position = 0
        filtered = (self.file_filter(
                    files,
                    pattern_lst[self.p.navi_position]))
        if filtered != []:
            filtered = sorted([dirname + os.sep + f for f in filtered])
            self.tkvars['fnames'].set(filtered)
            self.get_settings()

        # try next filter, if result is empty
        else:
            self.navigate(direction)

        # update file count
        self.num_label.config(text=len(self.p['fnames']))

    def file_filter(self, files, pattern):
        '''Filter a list of files to  match a pattern.

        Parameters
        ----------
        files : str[]
            A list of pathnames.
        pattern : str
            A regular expression for filtering the list.

        Returns
        -------
        str[]
            List items that match the pattern.
        '''
        filtered = []
        self.filter_hint.config(text=pattern)
        print('file filter: ' + pattern)
        p = re.compile(pattern)
        for f in files:
            if p.search(f):
                filtered.append(f)
        return filtered

    def __init_text_area(self, key):
        """
            Init a text area, here used as a lab-book, for example.

            The content is saved automatically to the parameter object,
            when the mouse leaves the text area.
        """
        self.ta.update({key: (tk.Text(self.set_frame[-1], undo=True))})
        ta = self.ta[list(self.ta)[-1]]
        ta.pack()
        ta.bind('<Leave>',
                (lambda _: self.__get_text(key, ta)))
        ttk.Button(self.set_frame[-1],
                   text='clear',
                   command=lambda: ta.delete(
            '1.0', tk.END)
        ).pack(fill='x')
        ttk.Button(self.set_frame[-1],
                   text='undo',
                   command=lambda: ta.edit_undo()
                   ).pack(fill='x')
        ttk.Button(self.set_frame[-1],
                   text='redo',
                   command=lambda: ta.edit_redo()
                   ).pack(fill='x')

    def __get_text(self, key, text_area):
        """Get text from text_area and copy it to parameter object."""
        self.p[key] = text_area.get('1.0', tk.END)

    def __listbox_selection_changed(self, event):
        """Handles selection change events of the file listbox."""
        try:
            self.index = event.widget.curselection()[0]
        except IndexError:
            pass  # nothing selected
        else:
            self.get_settings()
            self.show(self.p['fnames'][self.index])
            if self.p['data_information']:
                self.show_informations(self.p['fnames'][self.index])

    def __init_labelframe(self, key):
        """Add a label frame for widgets."""
        f = ttk.Frame(self.set_frame[-1])
        self.pane = ttk.Panedwindow(f, orient='vertical', width=400)
        self.lf = tk.LabelFrame(self.pane, text=self.p.label[key])
        self.lf.config(borderwidth=2, width=400, relief='groove')
        self.pane.add(self.lf)
        self.pane.pack(side='left', fill='both')
        f.pack(fill='both')

    def __init_sub_labelframe(self, key):
        """Add a label frame for widgets."""
        self.sub_lf = tk.LabelFrame(self.lf, text=self.p.label[key])
        self.sub_lf.config(borderwidth=2, width=400, relief='groove')
        self.sub_lf.pack(fill='both', pady=4, padx=4)

    def __init_post_button(self, event):
        f = ttk.Frame(self.lf)
        f.pack(fill='both')
        ttk.Button(f,
                   text='start postprocessing',
                   command=self.start_postprocessing).pack(side='top')

    def __init_horizontal_spacer(self, key):
        """Add a horizontal spacer line for widgets."""
        f = ttk.Frame(self.lf)
        hs = ttk.Separator(f)
        hs.pack(fill='x')
        f.pack(fill='both')

    def __init_sub_horizontal_spacer(self, key):
        """Add a horizontal spacer line for widgets"""
        f = ttk.Frame(self.sub_lf)
        hs = ttk.Separator(f)
        hs.pack(fill='x')
        f.pack(fill='both')

    def __init_label(self, key):
        f = ttk.Frame(self.lf)
        label1 = ttk.Label(f,
                           text=self.p.label[key])
        label1.pack(side='left')
        f.pack()

    def __init_entry(self, key):
        """
            Creates a label and an entry in a frame.

            A corresponding tk background textvariable is also crated. An
            option menu is created instead of en entry, if a hint is given
            in the parameter object. The help string in the parameter object
            is used for creating a tooltip.

            Parameter
            ---------
            key : str
                Key of a parameter obj.
        """
        padding = 2
        # sub label frames
        if(self.p.type[key] == 'sub_int' or
                self.p.type[key] == 'sub_float' or
                self.p.type[key] == 'sub'):
            f = ttk.Frame(self.sub_lf)
            f.pack(fill='x')
            l = ttk.Label(f, text=self.p.label[key])
            CreateToolTip(l, self.p.help[key])
            l.pack(side='left', padx=padding, pady=padding)
            if self.p.type[key] == 'sub_int':
                self.tkvars.update({key: tk.IntVar()})
            elif self.p.type[key] == 'sub_float':
                self.tkvars.update({key: tk.DoubleVar()})
            elif self.p.type[key] == 'sub':
                self.tkvars.update({key: tk.StringVar()})
            if self.p.hint[key] is not None:
                e = ttk.OptionMenu(f,
                                   self.tkvars[key],
                                   'spacer', *self.p.hint[key])
            else:
                e = ttk.Entry(f, width=17)
                e['textvariable'] = self.tkvars[key]
            CreateToolTip(e, self.p.help[key])
            e.pack(side='right', padx=padding, pady=padding)

        else:
            f = ttk.Frame(self.lf)
            f.pack(fill='x')
            l = ttk.Label(f, text=self.p.label[key])
            CreateToolTip(l, self.p.help[key])
            l.pack(side='left', padx=padding, pady=padding)
            if self.p.type[key] == 'int':
                self.tkvars.update({key: tk.IntVar()})
            elif self.p.type[key] == 'float':
                self.tkvars.update({key: tk.DoubleVar()})
            else:
                self.tkvars.update({key: tk.StringVar()})
            if self.p.hint[key] is not None:
                e = ttk.OptionMenu(f,
                                   self.tkvars[key],
                                   'spacer', *self.p.hint[key])
            else:
                e = ttk.Entry(f, width=17)
                e['textvariable'] = self.tkvars[key]
            CreateToolTip(e, self.p.help[key])
            e.pack(side='right', padx=padding, pady=padding)

    def __init_checkbutton(self, key):
        """Create a checkbutton with label and tooltip."""
        f = ttk.Frame(self.lf)
        f.pack(fill='x')
        self.tkvars.update({key: tk.BooleanVar()})
        self.tkvars[key].set(bool(self.p[key]))
        cb = ttk.Checkbutton(f)
        cb['variable'] = self.tkvars[key]
        cb['onvalue'] = True
        cb['offvalue'] = False
        cb['text'] = self.p.label[key]
        CreateToolTip(cb, self.p.help[key])
        cb.pack(side='left')

    def __init_sub_checkbutton(self, key):
        """Create a checkbutton with label and tooltip."""
        f = ttk.Frame(self.sub_lf)
        f.pack(fill='x')
        self.tkvars.update({key: tk.BooleanVar()})
        self.tkvars[key].set(bool(self.p[key]))
        cb = ttk.Checkbutton(f)
        cb['variable'] = self.tkvars[key]
        cb['onvalue'] = True
        cb['offvalue'] = False
        cb['text'] = self.p.label[key]
        CreateToolTip(cb, self.p.help[key])
        cb.pack(side='left')

    def __init_vec_colorpicker(self, key):
        whitespace = '                                 '
        f = ttk.Frame(self.lf)
        l = ttk.Label(f, text='invalid vector color')
        CreateToolTip(l, self.p.help[key])
        l.pack(side='left')
        self.invalid_color = tk.Button(f,
                                       text=whitespace,
                                       bg=self.p['invalid_color'],
                                       relief='groove',
                                       command=self.invalid_colorpicker)
        self.invalid_color.pack(side='right')
        f.pack(fill='x')

        f = ttk.Frame(self.lf)
        l = ttk.Label(f, text='valid vector color')
        CreateToolTip(l, self.p.help[key])
        l.pack(side='left')
        print(self.p['valid_color'])
        self.valid_color = tk.Button(f,
                                     text=whitespace,
                                     bg=self.p['valid_color'],
                                     relief='groove',
                                     command=self.valid_colorpicker)
        self.valid_color.pack(side='right')
        f.pack(fill='x')

    def invalid_colorpicker(self):
        color = colorchooser.askcolor()[1]
        if color is not None:
            self.p['invalid_color'] = color
            self.invalid_color.config(bg=self.p['invalid_color'])

    def valid_colorpicker(self):
        color = colorchooser.askcolor()[1]
        if color is not None:
            self.p['valid_color'] = color
            self.valid_color.config(bg=self.p['valid_color'])

    def log(self, columninformation=None, timestamp=False, text=None,
            group=None):
        """
            Add an entry to the lab-book.

            The first initialized text-area is assumed to be the
            lab-book.
            It is internally accessible by self.ta[0].

            Parameters
            ----------
            timestamp : bool
                Print current time.
                Pattern: yyyy-mm-dd hh:mm:ss.
                (default: False)
            text : str
                Print a text, a linebreak is appended.
                (default None)
            group : int
                Print group of parameters.
                (e.g. OpenPivParams.PIVPROC)
            columninformation : list
                Print column information of the selected file.

            Example
            -------
            log(text='processing parameters:',
                group=OpenPivParams.POSTPROC)
        """
        if text is not None:
            self.ta["lab_book_content"].insert(tk.END, text + '\n')
        if timestamp:
            td = datetime.today()
            s = '-'.join((str(td.year), str(td.month), str(td.day))) + \
                ' ' + \
                ':'.join((str(td.hour), str(td.minute), str(td.second)))
            self.log(text=s)
        if group is not None:
            self.log(text='Parameters:')
            for key in self.p.param:
                key_type = self.p.type[key]
                if key_type not in ['labelframe', 'sub_labelframe', 'h-spacer',
                                    'sub_h-spacer', 'post_button']:
                    if group < self.p.index[key] < group + 1000:
                        s = key + ': ' + str(self.p[key])
                        self.log(text=s)
        if columninformation is not None:
            self.ta["lab_book_content"]\
                .insert(tk.END, str(columninformation) + '\n')

    def show_informations(self, fname):
        """
            Shows the column names of the chosen file in the labbook.

            Parameters
            ----------
            fname : str
                A filename.
        """
        data = self.load_pandas(fname)
        if isinstance(data, str):
            self.log(text=data)
        else:
            self.log(columninformation=list(data.columns.values))

    def get_settings(self):
        """Copy widget variables to the parameter object."""
        for key in self.tkvars:
            if self.p.type[key] == 'str[]':
                self.p[key] = str2list(self.tkvars[key].get())
            elif self.p.type[key] == '[]':
                self.p[key] = self.p[key]
            else:
                self.p[key] = self.tkvars[key].get()
        for key in self.ta:
            self.__get_text(key, self.ta[key])

    def set_settings(self):
        """Copy values of the parameter object to widget variables."""
        for key in self.tkvars:
            self.tkvars[key].set(self.p[key])
        for key in self.ta:
            self.ta[key].delete('1.0', tk.END)
            self.ta[key].insert('1.0', self.p[key])

    def select_image_files(self):
        """Show a file dialog to select one or more filenames."""
        print('Use Ctrl + Shift to select multiple files.')
        files = filedialog.askopenfilenames(multiple=True)
        if len(files) > 0:
            self.p['fnames'] = list(files)
            self.tkvars['fnames'].set(self.p['fnames'])

        # update file count
        self.num_label.config(text=len(self.p['fnames']))

    def open_directory(self):
        """Show a dialog for opening a directory."""
        dir = filedialog.askdirectory()
        if len(dir) > 0:
            files = [dir + os.sep + file for file in os.listdir(dir)]
            self.p['fnames'] = list(files)
            self.tkvars['fnames'].set(self.p['fnames'])
        self.navigate('back')

        # update file count
        self.num_label.config(text=len(self.p['fnames']))

    def show(self, fname):
        """
            Display a file.

            This method distinguishes vector data (file extensions
            txt, dat, jvc,vec and csv) and images (all other file extensions).

            Parameters
            ----------
            fname : str
                A filename.
        """
        ext = fname.split('.')[-1]
        self.fig.clear()
        data = self.load_pandas(fname)
        if ext in ['txt', 'dat', 'jvc', 'vec', 'csv']:
            if self.p['plot_type'] == 'vectors':
                vec_plot.vector(
                    data,
                    self.p,
                    self.fig,
                    invert_yaxis=self.p['invert_yaxis'],
                    scale=self.p['vec_scale'],
                    width=self.p['vec_width'],
                    valid_color=self.p['valid_color'],
                    invalid_color=self.p['invalid_color']
                )
            elif self.p['plot_type'] == 'profiles':
                vec_plot.profiles(data, self.p,
                                  fname,
                                  self.fig,
                                  orientation=self.p['profiles_orientation']
                                  )
            elif self.p['plot_type'] == 'scatter':
                vec_plot.scatter(data,
                                 self.fig
                                 )
            elif self.p['plot_type'] == 'contour':
                vec_plot.contour(data, self.p, self.fig)

            elif self.p['plot_type'] == 'contour + vectors':
                vec_plot.contour_and_vector(data, self.p,
                                            self.fig,
                                            scale=self.p['vec_scale'],
                                            width=self.p['vec_width'])
            elif self.p['plot_type'] == 'streamlines':
                vec_plot.streamlines(data, self.p, self.fig)

            else:
                vec_plot.pandas_plot(data, self.p, self.fig)

        else:
            self.show_img(fname)
        self.fig.canvas.draw()

    def show_img(self, fname):
        """
            Display an image.

            Parameters
            ----------
            fname : str
            Pathname of an image file.
        """
        img = piv_tls.imread(fname)
        print('\nimage data type: {}'.format(img.dtype))
        print('max count: {}'.format(img.max()))
        print('min count {}:'.format(img.min()))
        if 'int' not in str(img.dtype):
            print('Warning: For PIV processing, ' +
                  'image will be normalized and converted to uint8. ' +
                  'This may cause a loss of precision.')

        print('Processing image.')
        img = img.astype(np.int32)
        # generate background if needed
        if self.p['background_subtract'] and \
                self.p['background_type'] != 'minA - minB':
            background = gen_background(self.p)

        elif self.p['background_subtract'] and \
                self.p['background_type'] == 'minA - minB':
            if fname == self.p['fnames'][-1]:
                img2 = self.p['fnames'][-2]
                img2 = piv_tls.imread(img2)
                background = gen_background(self.p, img2, img)
            else:
                img2 = self.p['fnames'][self.index + 1]
                img2 = piv_tls.imread(img2)
                background = gen_background(self.p, img, img2)
        else:
            background = None
        # preprocessing method became parameter due to the Add_In_Handler
        img = process_images(self, img, self.preprocessing_methods,
                             background=background)
        img = img.astype(np.int32)

        print('Processed image.')
        print('max count: {}'.format(img.max()))
        print('min count {}:'.format(img.min()))

        self.fig.add_subplot(111).matshow(img, cmap=plt.cm.Greys_r,
                                          vmax=self.p['matplot_intensity'])
        self.fig.canvas.draw()

    def get_parameters(self):
        return self.p

    def destroy_for_new_addins(self) -> None:
        """
            used to destroy the active object of the GUI and reinitialize
            it for inserting the addin methods and buttons

            :return: None
        """
        print('Updating GUI')
        # save old gui status
        self.p.dump_settings(self.p.params_fname)
        # destroy old gui
        tk.Tk.destroy(self)
        # initialize a new one
        self.__init__()

    def destroy(self):
        """Destroy the OpenPIV GUI.

        Settings are automatically saved.
        """
        if messagebox.askyesno(
            'Exit Manager',
                'Are you sure you want to exit?'):
            print('Saving settings')
            self.get_settings()
            self.p.dump_settings(self.p.params_fname)
            print('Destroying GUI')
            tk.Tk.destroy(self)
            # sometimes the GUI closes, but the main thread still runs
            print('Destorying main thread')
            sys.exit()
            print('Destroyed main thread.')
            # This should not execute if the thread is destroyed.
            # Could cause possible issue in the future.


if __name__ == '__main__':
    openPivGui = OpenPivGui()
    # a good starting size for the GUI
    openPivGui.geometry("1150x690")
    openPivGui.mainloop()



import tkinter as tk
from tkinter import ttk
import os
from functools import partial
import importlib

add_ins = {}
imported_add_ins = {}


def init_add_ins(gui):
    """
        In this method, the parameters that are already part of the GUI
        (loaded from OpenPIVParams) are extended by the parameters of
        the selected AddIns.
    :param gui: active instance of the OpenPivGui class
    :type gui: obj(OpenPivGui)
    :return:
    """
    # get the parameters which are already part of the gui
    parameters = gui.get_parameters()
    # get the add_ins loaded in the last session
    add_ins_to_be_included = parameters['used_addins']
    # Iterate through the selected add-ins, creating an instance of the
    # add-ins and reading out the variables of the class. These are then
    # appended to the parameter object.
    for add_in in add_ins_to_be_included:
        add_in_file = importlib.import_module("Add_Ins." + add_in)
        add_in_instance = getattr(add_in_file, add_in)(gui)
        imported_add_ins.update({add_in: add_in_instance.get_variables()})
        parameters.add_parameters(add_in_instance.get_variables())


def load_add_ins(gui) -> None:
    """
        Method that determines which add_ins are loaded on forced
        restart, and which parameters and methods must be removed based
         on the selection made.
        :param gui: active object of the OpenPivGui class
        :type gui: obj(OpenPivGui)
        :return: None
    """
    # get the add_ins loaded
    for add_in in add_ins:
        # if the add_in is chosen in the add_in menu
        if add_ins[add_in][0].get():
            imported_add_ins.update({add_in[:-3]: None})
            # if the add_in was not loaded yet
            if add_in[:-3] not in gui.get_parameters()['used_addins']:
                gui.get_parameters()['used_addins'].append(add_in[:-3])
        else:
            if add_in[:-3] in imported_add_ins.keys():
                imported_add_ins.pop(add_in[:-3])
                index = 0
                for i in gui.get_parameters()['used_addins']:
                    if i == add_in[:-3]:
                        gui.get_parameters()['used_addins'].pop(index)
                    index += 1
            if add_in[:-3] in gui.buttons:
                gui.buttons.pop(add_in[:-3])
            if add_in[:-3] in gui.preprocessing_methods:
                gui.preprocessing_methods.pop(add_in[:-3])
    # forced restart
    gui.destroy_for_new_addins()


def linebreak(text):
    """
        used to make the text fit in the textarea in the add_in menu
        :param text: add_in_tip which has to be cutted
        :return: add_in_tip within the linebreak needed
    """
    if len(text) > 60:
        char = 60
        while char > 0:
            if text[char] == " ":
                returntext = text[0:char] + "\n" + linebreak(text[char + 1:])
                break
            else:
                char = char - 1
        return returntext
    else:
        return text


def description(add_in: list) -> None:
    """
        method, which is used to load the add_in_tip from the specified
        add_in and then output it in the textarea
        :param add_in: list containing add_in name and an obj(OpenPivGui)
        :type add_in: list
        :return: None
    """
    # get add_in_py_file within the use of importlib
    add_in_py_file = importlib.import_module("Add_Ins." + add_in[0])
    # get the init method of the given add_in
    add_in_init = getattr(add_in_py_file, add_in[0])
    add_in_obj = add_in_init(add_in[1])
    # get the add_in_tip
    add_in_description = add_in_obj.get_description()
    add_in_description = linebreak(add_in_description)
    # print the add_in_tip in the textarea
    textarea.config(state="normal")
    textarea.delete("3.0", tk.END)
    textarea.insert("3.0", "\n" + add_in_description)
    textarea.config(state="disabled")


def select_add_ins(gui) -> None:
    """
        Creates a popup menu in which the user is able to decide which
        Add Ins he/she wants to be included and which he/she wants to remove
        By pressing the Save button the load_addins function which is used
        to store the selected addins is called.
        :param gui: active object of the OpenPivGui class
        :type gui: obj(OpenPivGui)
        :return: None
    """
    global add_in_tk_frame
    add_in_tk_frame = tk.Toplevel()
    add_in_tk_frame.geometry("850x690")
    selection_menu = ttk.Frame(add_in_tk_frame)
    selection_menu.pack(side='left',
                        fill='both',
                        expand='True')
    available_add_ins = os.listdir(os.path.dirname(__file__) + "/Add_Ins")
    possible_add_ins = ["general", "preprocessing",
                        "postprocessing", "other", "plotting"]
    buttons = []
    add_ins_chosen = gui.get_parameters()['used_addins']
    j = 1
    for i in possible_add_ins:
        tk.Label(selection_menu, text=i.capitalize(),
                 font='Helvetica 10 bold').pack(padx=5, side="top",
                                                fill="both")
        j += 1
        for add_in in available_add_ins:
            add_in_string_parts = add_in.split("_")
            if add_in_string_parts[-1][:-3] == i:
                add_ins.update({add_in: [tk.BooleanVar(),
                                         add_in_string_parts[-1][:-3],
                                         "information(add_in[:-3])"]})
                if add_in[:-3] in add_ins_chosen:
                    add_ins[add_in][0].set(True)
                else:
                    add_ins[add_in][0].set(False)
                tk.Checkbutton(selection_menu,
                               var=add_ins[add_in][0],
                               onvalue=True, offvalue=False)\
                    .pack(padx=5)
                buttons.append(tk.Button(selection_menu, text=add_in[:-3],
                                         command=partial(description, [add_in[:-3],
                                                                       gui])))
                buttons[-1].pack(padx=5)
                j += 1
        ttk.Separator(selection_menu).pack(side="top", expand=2)
    global textarea
    textarea = tk.Text(add_in_tk_frame)
    textarea.pack(side="left", fill="both")
    textarea.insert("1.0", "ADD_IN DESCRIPTION \n")
    textarea.insert(tk.END, 60 * "-" + "\n")
    textarea.insert(tk.END, "variable")
    textarea.config(state="disabled")

    tk.Button(selection_menu, text="Save", bg="white", command=lambda:
              load_add_ins(gui)) \
        .pack(side="bottom", fill="both", expand="True")



import setuptools
from glob import glob

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="openpivgui",
    version="0.4.5",
    install_requires=['OpenPiv', 'pandas'],
    author="P. Vennemann and contributors.",
    author_email="vennemann@fh-muenster.de",
    description="A simple GUI for Open PIV.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/OpenPIV/openpiv_tk_gui",
    include_package_data=True,
    package_data={'': ['./res/*.png']},
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.0',
)
